import numpy as np
import pytest

from glotaran.io import load_model
from glotaran.io import load_parameters
from glotaran.optimization.optimize import optimize
from glotaran.project import Scheme
from glotaran.simulation import simulate

MODEL_3C_BASE = """\
dataset:
    dataset1:
        megacomplex: [mc1]
        global_megacomplex: [mc2]
        initial_concentration: j1
        irf: irf1
    dataset2:
        megacomplex: [mc2]
        global_megacomplex: [mc1]
        initial_concentration: j1
        irf: irf1
    dataset3:
        megacomplex: [mc1]
        initial_concentration: j1
        irf: irf1
    dataset4:
        megacomplex: [mc2]
        initial_concentration: j1
megacomplex:
    mc1:
        type: decay
        k_matrix: [k1]
    mc2:
        type: spectral
        shape:
            s1: sh1
            s2: sh2
            s3: sh3
irf:
    irf1:
        type: multi-gaussian
        center: [irf.center]
        width: [irf.width]
shape:
    sh1:
        type: gaussian
        amplitude: shapes.amps.1
        location: shapes.locs.1
        width: shapes.width.1
    sh2:
        type: gaussian
        amplitude: shapes.amps.2
        location: shapes.locs.2
        width: shapes.width.2
    sh3:
        type: gaussian
        amplitude: shapes.amps.3
        location: shapes.locs.3
        width: shapes.width.3
"""

MODEL_3C_PARALLEL = f"""\
{MODEL_3C_BASE}
initial_concentration:
    j1:
        compartments: [s1, s2, s3]
        parameters: [j.1, j.1, j.1]
k_matrix:
    k1:
        matrix:
            (s1, s1): "kinetic.1"
            (s2, s2): "kinetic.2"
            (s3, s3): "kinetic.3"
"""

MODEL_3C_SEQUENTIAL = f"""\
{MODEL_3C_BASE}
initial_concentration:
    j1:
        compartments: [s1, s2, s3]
        parameters: [j.1, j.0, j.0]
k_matrix:
    k1:
        matrix:
            (s2, s1): "kinetic.1"
            (s3, s2): "kinetic.2"
            (s3, s3): "kinetic.3"
"""

PARAMETERS_3C_BASE = """\
irf:
    - ["center", 1.3]
    - ["width", 7.8]
j:
    - ["1", 1, {"vary": False, "non-negative": False}]
    - ["0", 0, {"vary": False, "non-negative": False}]
"""

PARAMETERS_3C_BASE_PARALLEL = f"""\
{PARAMETERS_3C_BASE}
shapes:
    amps: [7, 3, 30, {{"vary": False}}]
    locs: [620, 670, 720, {{"vary": False}}]
    width: [10, 30, 50, {{"vary": False}}]
"""

PARAMETERS_3C_BASE_SEQUENTIAL = f"""\
{PARAMETERS_3C_BASE}
shapes:
    amps:
        - 9
        - 7
        - 5
        - {{"vary": True, min: 0, max: 10}}
    locs: [610, 670, 730, {{"vary": False}}]
    width: [15, 25, 10, {{"vary": False}}]
"""

PARAMETERS_3C_PARALLEL_WANTED = f"""\
kinetic:
    - ["1", 301e-3]
    - ["2", 502e-4]
    - ["3", 705e-5]
{PARAMETERS_3C_BASE_PARALLEL}
"""

PARAMETERS_3C_INITIAL_PARALLEL = f"""\
kinetic:
    - ["1", 300e-3, {{non-negative: true}}]
    - ["2", 500e-4, {{non-negative: true}}]
    - ["3", 700e-5, {{non-negative: true}}]
{PARAMETERS_3C_BASE_PARALLEL}
"""

PARAMETERS_3C_SIM_SEQUENTIAL = f"""\
kinetic:
    - ["1", 501e-3, {{non-negative: true}}]
    - ["2", 202e-4, {{non-negative: true}}]
    - ["3", 105e-5, {{non-negative: true}}]
{PARAMETERS_3C_BASE_SEQUENTIAL}
"""

PARAMETERS_3C_INITIAL_SEQUENTIAL = f"""\
kinetic:
    - ["1", 500e-3]
    - ["2", 200e-4]
    - ["3", 100e-5]
    - {{"non-negative": True}}
{PARAMETERS_3C_BASE_SEQUENTIAL}
"""


class ThreeComponentParallel:
    model = load_model(MODEL_3C_PARALLEL, format_name="yml_str")
    initial_parameters = load_parameters(PARAMETERS_3C_INITIAL_PARALLEL, format_name="yml_str")
    wanted_parameters = load_parameters(PARAMETERS_3C_PARALLEL_WANTED, format_name="yml_str")
    time = np.arange(-10, 100, 1.5)
    spectral = np.arange(600, 750, 10)
    axis = {"time": time, "spectral": spectral}


class ThreeComponentSequential:
    model = load_model(MODEL_3C_SEQUENTIAL, format_name="yml_str")
    initial_parameters = load_parameters(PARAMETERS_3C_INITIAL_SEQUENTIAL, format_name="yml_str")
    wanted_parameters = load_parameters(PARAMETERS_3C_SIM_SEQUENTIAL, format_name="yml_str")
    time = np.arange(-10, 50, 1.0)
    spectral = np.arange(600, 750, 5.0)
    axis = {"time": time, "spectral": spectral}


@pytest.mark.parametrize(
    "suite",
    [
        ThreeComponentParallel,
        ThreeComponentSequential,
    ],
)
@pytest.mark.parametrize("nnls", [True, False])
def test_kinetic_model(suite, nnls):
    model = suite.model
    print(model.validate())
    assert model.valid()
    model.dataset_groups["default"].method = (
        "non_negative_least_squares" if nnls else "variable_projection"
    )

    wanted_parameters = suite.wanted_parameters
    print(model.validate(wanted_parameters))
    print(wanted_parameters)
    assert model.valid(wanted_parameters)

    initial_parameters = suite.initial_parameters
    print(model.validate(initial_parameters))
    assert model.valid(initial_parameters)

    print(model.markdown(wanted_parameters))

    dataset = simulate(model, "dataset1", wanted_parameters, suite.axis)

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["spectral"].size)

    data = {f"dataset{i}": dataset for i in range(1, 5)}

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for param in result.optimized_parameters.all():
        assert np.allclose(param.value, wanted_parameters.get(param.label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]

    print(resultdata)

    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["spectral"], resultdata["spectral"])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, rtol=1e-2)

    # 3 compartments * 3 shapes (dataset1)
    # + 3 shapes * 3 compartments (dataset2)
    expected_number_of_clps_common = 18
    if suite is ThreeComponentParallel:
        # + 3 compartments * 15 items in global axis (dataset3)
        # + 3 shapes * 74 items in global axis (dataset4)
        assert result.number_of_clps == expected_number_of_clps_common + 45 + 222
    else:
        # + 3 compartments * 30 items in global axis (dataset3)
        # + 3 shapes * 60 items in global axis (dataset4)
        assert result.number_of_clps == expected_number_of_clps_common + 90 + 180



"""Tests for ``glotaran.project.optimization_history``."""

from pathlib import Path
from textwrap import dedent

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pandas.testing import assert_series_equal

from glotaran.optimization.optimization_history import OptimizationHistory


def test_optimization_history_init_no_data():
    """Empty DataFrame with correct columns and index."""
    result = OptimizationHistory()

    assert result.shape == (0, 5)
    assert all(result.columns == ["nfev", "cost", "cost_reduction", "step_norm", "optimality"])
    assert result.index.name == "iteration"


@pytest.mark.parametrize(
    "optimize_stdout, expected_df",
    (
        (
            "random string",
            pd.DataFrame(
                None,
                columns=["iteration", "nfev", "cost", "cost_reduction", "step_norm", "optimality"],
            ).set_index("iteration"),
        ),
        (
            dedent(
                """\
                   Iteration     Total nfev        Cost      Cost reduction    Step norm     Optimality
                       0              1         7.5834e+00                                    3.84e+01
                       1              2         7.5833e+00      1.37e-04       4.55e-05       1.26e-01
                       2              3         7.5833e+00      6.02e-11       6.44e-09       1.64e-05
                Both `ftol` and `xtol` termination conditions are satisfied.
                Function evaluations 3, initial cost 7.5834e+00, final cost 7.5833e+00, first-order optimality 1.64e-05.
                """  # noqa: E501
            ),
            pd.DataFrame(
                {
                    "iteration": [0, 1, 2],
                    "nfev": [1, 2, 3],
                    "cost": [7.5834, 7.5833, 7.5833],
                    "cost_reduction": [np.nan, 1.37e-4, 6.02e-11],
                    "step_norm": [np.nan, 4.55e-5, 6.44e-9],
                    "optimality": [38.4, 0.126, 1.64e-5],
                },
                columns=["iteration", "nfev", "cost", "cost_reduction", "step_norm", "optimality"],
            ).set_index("iteration"),
        ),
    ),
)
def test_optimization_history_init_from_stdout_str(
    optimize_stdout: str, expected_df: pd.DataFrame, tmp_path: Path
):
    """Parsing from ``optimize_stdout`` behaves gracefully and creates expected DataFrame."""
    result = OptimizationHistory.from_stdout_str(optimize_stdout)

    assert result.shape == expected_df.shape
    assert all(result.columns == ["nfev", "cost", "cost_reduction", "step_norm", "optimality"])
    assert result.index.name == "iteration"

    assert_frame_equal(result.data, expected_df)

    # pandas like access
    assert_series_equal(result.cost, expected_df.cost)
    assert_series_equal(result["cost"], expected_df["cost"])

    assert hasattr(result, "plot")

    # round tripping
    save_path = tmp_path / "optimization_history.csv"
    result.to_csv(save_path)

    round_tripped = OptimizationHistory.from_csv(save_path)

    assert_frame_equal(result._df, round_tripped._df)
    assert round_tripped.source_path == save_path.as_posix()
    assert result.source_path == round_tripped.source_path



"""Module containing the optimizer class."""
from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
from scipy.optimize import OptimizeResult
from scipy.optimize import least_squares

from glotaran import __version__ as glotaran_version
from glotaran.optimization.optimization_group import OptimizationGroup
from glotaran.optimization.optimization_history import OptimizationHistory
from glotaran.parameter import ParameterHistory
from glotaran.project import Result
from glotaran.project import Scheme
from glotaran.utils.regex import RegexPattern
from glotaran.utils.tee import TeeContext

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike

SUPPORTED_METHODS = {
    "TrustRegionReflection": "trf",
    "Dogbox": "dogbox",
    "Levenberg-Marquardt": "lm",
}


class InitialParameterError(ValueError):
    """Indicates that initial parameters can not be evaluated."""

    def __init__(self):
        """Initialize a InitialParameterError."""
        super().__init__("Initial parameters can not be evaluated.")


class ParameterNotInitializedError(ValueError):
    """Indicates that scheme parameters are not initialized."""

    def __init__(self):
        """Initialize a ParameterNotInitializedError."""
        super().__init__("Parameter not initialized")


class MissingDatasetsError(ValueError):
    """Indicates that datasets are missing in the scheme."""

    def __init__(self, missing_datasets: list[str]):
        """Initialize a MissingDatasetsError.

        Parameters
        ----------
        missing_datasets : list[str]
            The missing datasets.
        """
        super().__init__(f"Missing data for datasets: {missing_datasets}")


class UnsupportedMethodError(ValueError):
    """Indicates that the optimization method is unsupported."""

    def __init__(self, method: str):
        """Initialize an UnsupportedMethodError.

        Parameters
        ----------
        method : str
            The unsupported method.
        """
        super().__init__(
            f"Unsupported optimization method {method}. "
            f"Supported methods are '{list(SUPPORTED_METHODS.keys())}'"
        )


class Optimizer:
    """A class to optimize a scheme."""

    def __init__(self, scheme: Scheme, verbose: bool = True, raise_exception: bool = False):
        """Initialize an optimization group for a dataset group.

        Parameters
        ----------
        scheme : Scheme
            The optimization scheme.
        verbose : bool
            Deactivate printing of logs if `False`.
        raise_exception : bool
            Raise exceptions during optimizations instead of gracefully exiting if `True`.

        Raises
        ------
        MissingDatasetsError
            Raised if datasets are missing.
        ParameterNotInitializedError
            Raised if the scheme parameters are `None`.
        UnsupportedMethodError
            Raised if the optimization method is unsupported.
        """
        if missing_datasets := [
            label for label in scheme.model.dataset if label not in scheme.data
        ]:
            raise MissingDatasetsError(missing_datasets)
        if scheme.parameters is None:
            raise ParameterNotInitializedError()
        self._parameters = scheme.parameters.copy()
        if scheme.optimization_method not in SUPPORTED_METHODS:
            raise UnsupportedMethodError(scheme.optimization_method)
        self._method = SUPPORTED_METHODS[scheme.optimization_method]

        self._scheme = scheme
        self._tee = TeeContext()
        self._verbose = verbose
        self._raise = raise_exception

        self._optimization_result: OptimizeResult = None
        self._termination_reason = ""

        self._optimization_groups = [
            OptimizationGroup(scheme, group)
            for group in scheme.model.get_dataset_groups().values()
        ]

        self._parameter_history = ParameterHistory()
        self._parameter_history.append(scheme.parameters)

    def optimize(self):
        """Perform the optimization.

        Raises
        ------
        Exception
            Raised if an exception occurs during optimization and raise_exception is `True`.
        """
        (
            self._free_parameter_labels,
            initial_parameter,
            lower_bounds,
            upper_bounds,
        ) = self._scheme.parameters.get_label_value_and_bounds_arrays(exclude_non_vary=True)
        with self._tee:
            try:
                verbose = 2 if self._verbose else 0
                self._optimization_result = least_squares(
                    self.objective_function,
                    initial_parameter,
                    bounds=(lower_bounds, upper_bounds),
                    method=self._method,
                    max_nfev=self._scheme.maximum_number_function_evaluations,
                    verbose=verbose,
                    ftol=self._scheme.ftol,
                    gtol=self._scheme.gtol,
                    xtol=self._scheme.xtol,
                )
                self._termination_reason = self._optimization_result.message
            except Exception as e:
                if self._raise:
                    raise e
                warn(f"Optimization failed:\n\n{e}")
                self._termination_reason = str(e)

    def objective_function(self, parameters: ArrayLike) -> ArrayLike:
        """Calculate the objective for the optimization.

        Parameters
        ----------
        parameters : ArrayLike
            the parameters provided by the optimizer.

        Returns
        -------
        ArrayLike
            The objective for the optimizer.
        """
        self._parameters.set_from_label_and_value_arrays(self._free_parameter_labels, parameters)
        return self.calculate_penalty()

    def calculate_penalty(self) -> ArrayLike:
        """Calculate the penalty of the scheme.

        Returns
        -------
        ArrayLike
            The penalty.
        """
        for group in self._optimization_groups:
            group.calculate(self._parameters)
        self._parameter_history.append(
            self._parameters, self.get_current_optimization_iteration(self._tee.read())
        )

        penalties = [group.get_full_penalty() for group in self._optimization_groups]

        return np.concatenate(penalties) if len(penalties) != 1 else penalties[0]

    def create_result(self) -> Result:
        """Create the result of the optimization.

        Returns
        -------
        Result
            The result of the optimization.

        Raises
        ------
        InitialParameterError
            Raised if the initial parameters could not be evaluated.
        """
        success = self._optimization_result is not None

        if self._parameter_history.number_of_records == 1:
            raise InitialParameterError()
        elif not success:
            self._parameters.set_from_history(self._parameter_history, -2)

        result_args = {
            "success": success,
            "scheme": self._scheme,
            "glotaran_version": glotaran_version,
            "free_parameter_labels": self._free_parameter_labels,
            "initial_parameters": self._scheme.parameters,
            "parameter_history": self._parameter_history,
            "termination_reason": self._termination_reason,
            "optimization_history": OptimizationHistory.from_stdout_str(self._tee.read()),
            "number_of_function_evaluations": self._optimization_result.nfev
            if success
            else self._parameter_history.number_of_records,
        }

        if success:
            result_args["number_of_jacobian_evaluations"] = self._optimization_result.njev
            result_args["optimality"] = float(self._optimization_result.optimality)
            result_args["number_of_residuals"] = self._optimization_result.fun.size
            result_args["number_of_clps"] = sum(
                group.number_of_clps for group in self._optimization_groups
            )
            result_args["number_of_free_parameters"] = self._optimization_result.x.size
            result_args["degrees_of_freedom"] = (
                result_args["number_of_residuals"]
                - result_args["number_of_free_parameters"]
                - result_args["number_of_clps"]
            )
            result_args["chi_square"] = float(np.sum(self._optimization_result.fun**2))
            result_args["reduced_chi_square"] = (
                result_args["chi_square"] / result_args["degrees_of_freedom"]
            )
            result_args["root_mean_square_error"] = float(
                np.sqrt(result_args["reduced_chi_square"])
            )
            result_args["jacobian"] = self._optimization_result.jac

            self._parameters.set_from_label_and_value_arrays(
                self._free_parameter_labels, self._optimization_result.x
            )
            result_args[
                "covariance_matrix"
            ] = self.calculate_covariance_matrix_and_standard_errors(
                result_args["jacobian"], result_args["root_mean_square_error"]
            )

        result_args["additional_penalty"] = [
            group.get_additional_penalties() for group in self._optimization_groups
        ]

        full_penalty = self.calculate_penalty()
        result_args["cost"] = 0.5 * np.dot(full_penalty, full_penalty)

        result_args["optimized_parameters"] = self._parameters

        result_args["data"] = {}
        for group in self._optimization_groups:
            group.calculate(self._parameters)
            result_args["data"].update(group.create_result_data())

        return Result(**result_args)

    def calculate_covariance_matrix_and_standard_errors(
        self, jacobian: ArrayLike, root_mean_square_error: float
    ) -> ArrayLike:
        """Calculate the covariance matrix and standard errors of the optimization.

        Parameters
        ----------
        jacobian : ArrayLike
            The jacobian matrix.
        root_mean_square_error : float
            The root mean square error.

        Returns
        -------
        ArrayLike
            The covariance matrix.
        """
        # See PR #706: More robust covariance matrix calculation
        _, jacobian_sv, jacobian_rsv = np.linalg.svd(jacobian, full_matrices=False)
        jacobian_sv_square = jacobian_sv**2
        mask = jacobian_sv_square > np.finfo(float).eps
        covariance_matrix = (jacobian_rsv[mask].T / jacobian_sv_square[mask]) @ jacobian_rsv[mask]
        standard_errors = root_mean_square_error * np.sqrt(np.diag(covariance_matrix))
        for label, error in zip(self._free_parameter_labels, standard_errors):
            self._parameters.get(label).standard_error = error
        return covariance_matrix

    @staticmethod
    def get_current_optimization_iteration(optimize_stdout: str) -> int:
        """Extract current iteration from ``optimize_stdout``.

        Parameters
        ----------
        optimize_stdout: str
            SciPy optimization stdout string, read out via ``TeeContext.read()``.

        Returns
        -------
        int
            Current iteration (``0`` if pattern did not match).
        """
        matches = RegexPattern.optimization_stdout.findall(optimize_stdout)
        return 0 if len(matches) == 0 else int(matches[-1][0])



"""This module contains the megacomplex."""
from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING
from typing import ClassVar

import xarray as xr
from attrs import NOTHING
from attrs import fields

from glotaran.model.item import ModelItemTyped
from glotaran.model.item import item
from glotaran.plugin_system.megacomplex_registration import register_megacomplex

if TYPE_CHECKING:
    from glotaran.model import DatasetModel
    from glotaran.typing.types import ArrayLike


def megacomplex(
    *,
    dataset_model_type: type[DatasetModel] | None = None,
    exclusive: bool = False,
    unique: bool = False,
) -> Callable:
    """Create a megacomplex from a class.

    Parameters
    ----------
    dataset_model_type: type
        The dataset model type.
    exclusive: bool
        Whether the megacomplex is exclusive.
    unique: bool
        Whether the megacomplex is unique.

    Returns
    -------
    Callable
    """

    def decorator(cls):
        megacomplex_type = item(cls)
        megacomplex_type.__dataset_model_type__ = dataset_model_type
        megacomplex_type.__is_exclusive__ = exclusive
        megacomplex_type.__is_unique__ = unique

        megacomplex_type_str = fields(cls).type.default
        if megacomplex_type_str is not NOTHING:
            register_megacomplex(megacomplex_type_str, megacomplex_type)

        return megacomplex_type

    return decorator


@item
class Megacomplex(ModelItemTyped):
    """A base class for megacomplex models.

    Subclasses must overwrite :method:`glotaran.model.Megacomplex.calculate_matrix`
    and :method:`glotaran.model.Megacomplex.index_dependent`.
    """

    dimension: str | None = None

    __dataset_model_type__: ClassVar[type | None] = None
    __is_exclusive__: ClassVar[bool]
    __is_unique__: ClassVar[bool]

    @classmethod
    def get_dataset_model_type(cls) -> type | None:
        """Get the dataset model type.

        Returns
        -------
        type | None
        """
        return cls.__dataset_model_type__

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ) -> tuple[list[str], ArrayLike]:
        """Calculate the megacomplex matrix.

        Parameters
        ----------
        dataset_model: DatasetModel
            The dataset model.
        global_axis: ArrayLike
            The global axis.
        model_axis: ArrayLike
            The model axis.
        **kwargs
            Additional arguments.

        Returns
        -------
        tuple[list[str], ArrayLike]:
            The clp labels and the matrix.

        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        """Finalize a dataset.

        Parameters
        ----------
        dataset_model: DatasetModel
            The dataset model.
        dataset: xr.Dataset
            The dataset.
        is_full_model: bool
            Whether the model is a full model.
        as_global: bool
            Whether megacomplex is calculated as global megacomplex.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError


def is_exclusive(cls: type[Megacomplex]) -> bool:
    """Check if the megacomplex is exclusive.

    Parameters
    ----------
    cls: type[Megacomplex]
        The megacomplex type.

    Returns
    -------
    bool
    """
    return cls.__is_exclusive__


def is_unique(cls: type[Megacomplex]) -> bool:
    """Check if the megacomplex is unique.

    Parameters
    ----------
    cls: type[Megacomplex]
        The megacomplex type.

    Returns
    -------
    bool
    """
    return cls.__is_unique__



"""The glotaran parameter registry module."""
from __future__ import annotations

from pathlib import Path
from typing import Literal

from glotaran.builtin.io.yml.utils import write_dict
from glotaran.io import load_parameters
from glotaran.io import save_parameters
from glotaran.model import Model
from glotaran.parameter import Parameters
from glotaran.plugin_system.project_io_registration import supported_file_extensions_project_io
from glotaran.project.project_registry import ProjectRegistry


class ProjectParameterRegistry(ProjectRegistry):
    """A registry for parameters."""

    def __init__(self, directory: Path):
        """Initialize a parameter registry.

        Parameters
        ----------
        directory : Path
            The registry directory.
        """
        super().__init__(
            directory / "parameters",
            supported_file_extensions_project_io("load_parameters"),
            load_parameters,
            item_name="Parameters",
        )

    def generate_parameters(
        self,
        model: Model,
        name: str | None,
        *,
        format_name: Literal["yml", "yaml", "csv"] = "csv",
        allow_overwrite: bool = False,
        ignore_existing: bool = False,
    ):
        """Generate parameters for a model.

        Parameters
        ----------
        model : Model
            The model.
        name : str | None
            The name of the parameters.
        format_name : Literal["yml", "yaml", "csv"]
            The parameter format.
        allow_overwrite: bool
            Whether to overwrite existing parameters.
        ignore_existing: bool
            Whether to ignore generation of a parameter file if it already exists.

        Raises
        ------
        FileExistsError
            Raised if parameters is already existing and `allow_overwrite=False`.
        """
        parameters = model.generate_parameters().to_parameter_dict_or_list(
            serialize_parameters=True
        )
        parameter_file = self.directory / f"{name}.{format_name}"

        if parameter_file.exists() and ignore_existing:
            return

        if parameter_file.exists() and not allow_overwrite:
            raise FileExistsError(
                f"Parameters {name!r} already exists and `allow_overwrite=False`."
            )

        if format_name in ["yml", "yaml"]:
            write_dict(parameters, file_name=parameter_file, offset=0)
        elif format_name == "csv":
            save_parameters(
                Parameters.from_dict(parameters)
                if isinstance(parameters, dict)
                else Parameters.from_list(parameters),
                parameter_file,
                allow_overwrite=allow_overwrite,
            )



"""Data Io registration convenience functions.

Note
----
The [call-arg] type error would be raised since the base methods doesn't have a ``**kwargs``
argument, but we rather ignore this error here, than adding ``**kwargs`` to the base method
and causing an [override] type error in the plugins implementation.
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import xarray as xr
from tabulate import tabulate

from glotaran.io.interface import DataIoInterface
from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.base_registry import add_instantiated_plugin_to_registry
from glotaran.plugin_system.base_registry import get_method_from_plugin
from glotaran.plugin_system.base_registry import get_plugin_from_registry
from glotaran.plugin_system.base_registry import is_registered_plugin
from glotaran.plugin_system.base_registry import methods_differ_from_baseclass_table
from glotaran.plugin_system.base_registry import registered_plugins
from glotaran.plugin_system.base_registry import set_plugin
from glotaran.plugin_system.base_registry import show_method_help
from glotaran.plugin_system.base_registry import supported_file_extensions
from glotaran.plugin_system.io_plugin_utils import bool_table_repr
from glotaran.plugin_system.io_plugin_utils import infer_file_format
from glotaran.plugin_system.io_plugin_utils import not_implemented_to_value_error
from glotaran.plugin_system.io_plugin_utils import protect_from_overwrite
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Generator
    from collections.abc import Sequence
    from typing import Any
    from typing import Literal

    from glotaran.io.interface import DataLoader
    from glotaran.io.interface import DataSaver
    from glotaran.typing import StrOrPath

DATA_IO_METHODS = ("load_dataset", "save_dataset")
"""Methods used by DataIoInterface plugins."""


def register_data_io(
    format_names: str | list[str],
) -> Callable[[type[DataIoInterface]], type[DataIoInterface]]:
    """Register data io plugins to one or more formats.

    Decorate a data io plugin class with ``@register_data_io(format_name|[*format_names])``
    to add it to the registry.

    Parameters
    ----------
    format_names : str | list[str]
        Name of the data io plugin under which it is registered.

    Returns
    -------
    Callable[[type[DataIoInterface]], type[DataIoInterface]]
        Inner decorator function.

    Examples
    --------
    >>> @register_data_io("my_format_1")
    ... class MyDataIo1(DataIoInterface):
    ...     pass

    >>> @register_data_io(["my_format_1", "my_format_1_alias"])
    ... class MyDataIo2(DataIoInterface):
    ...     pass
    """

    def wrapper(cls: type[DataIoInterface]) -> type[DataIoInterface]:
        add_instantiated_plugin_to_registry(
            plugin_register_keys=format_names,
            plugin_class=cls,
            plugin_registry=__PluginRegistry.data_io,
            plugin_set_func_name="set_data_plugin",
        )
        return cls

    return wrapper


def is_known_data_format(format_name: str) -> bool:
    """Check if a data format is in the data_io registry.

    Parameters
    ----------
    format_name : str
        Name of the data io plugin under which it is registered.

    Returns
    -------
    bool
        Whether or not the data format is a registered data io plugins.
    """
    return is_registered_plugin(
        plugin_register_key=format_name, plugin_registry=__PluginRegistry.data_io
    )


def known_data_formats(full_names: bool = False) -> list[str]:
    """Names of the registered data io plugins.

    Parameters
    ----------
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    list[str]
        List of registered data io plugins.
    """
    return registered_plugins(plugin_registry=__PluginRegistry.data_io, full_names=full_names)


def set_data_plugin(
    format_name: str,
    full_plugin_name: str,
) -> None:
    """Set the plugin used for a specific data format.

    This function is useful when you want to resolve conflicts of installed plugins
    or overwrite the plugin used for a specific format.

    Effected functions:

    - :func:`load_dataset`
    - :func:`save_dataset`

    Parameters
    ----------
    format_name : str
        Format name used to refer to the plugin when used for ``save`` and ``load`` functions.
    full_plugin_name : str
        Full name (import path) of the registered plugin.
    """
    set_plugin(
        plugin_register_key=format_name,
        full_plugin_name=full_plugin_name,
        plugin_registry=__PluginRegistry.data_io,
    )


def get_data_io(format_name: str) -> DataIoInterface:
    """Retrieve a data io plugin from the data_io registry.

    Parameters
    ----------
    format_name : str
        Name of the data io plugin under which it is registered.

    Returns
    -------
    DataIoInterface
        Data io plugin instance.
    """
    return get_plugin_from_registry(
        plugin_register_key=format_name,
        plugin_registry=__PluginRegistry.data_io,
        not_found_error_message=(
            f"Unknown  Data Io format {format_name!r}. Known formats are: {known_data_formats()}"
        ),
    )


@not_implemented_to_value_error
def load_dataset(
    file_name: StrOrPath, format_name: str | None = None, **kwargs: Any
) -> xr.Dataset:
    """Read data from a file to :xarraydoc:`Dataset` or :xarraydoc:`DataArray`.

    Parameters
    ----------
    file_name : StrOrPath
        File containing the data.
    format_name : str
        Format the file is in, if not provided it will be inferred from the file extension.
    **kwargs : Any
        Additional keyword arguments passes to the ``read_dataset`` implementation
        of the data io plugin. If you aren't sure about those use ``get_dataloader``
        to get the implementation with the proper help and autocomplete.

    Returns
    -------
    xr.Dataset
        Data loaded from the file.
    """
    io = get_data_io(format_name or infer_file_format(file_name))
    dataset = io.load_dataset(Path(file_name).as_posix(), **kwargs)

    if isinstance(dataset, xr.DataArray):
        dataset = dataset.to_dataset(name="data")
    dataset.attrs["loader"] = load_dataset
    dataset.attrs["source_path"] = Path(file_name).as_posix()
    return dataset


@not_implemented_to_value_error
def save_dataset(
    dataset: xr.Dataset | xr.DataArray,
    file_name: StrOrPath,
    format_name: str | None = None,
    *,
    data_filters: list[str] | None = None,
    allow_overwrite: bool = False,
    update_source_path: bool = True,
    **kwargs: Any,
) -> None:
    """Save data from :xarraydoc:`Dataset` or :xarraydoc:`DataArray` to a file.

    Parameters
    ----------
    dataset : xr.Dataset | xr.DataArray
        Data to be written to file.
    file_name : StrOrPath
        File to write the data to.
    format_name : str
        Format the file should be in, if not provided it will be inferred from the file extension.
    data_filters : list[str] | None
        Optional list of items in the dataset to be saved.
    allow_overwrite : bool
        Whether or not to allow overwriting existing files, by default False
    update_source_path: bool
        Whether or not to update the ``source_path`` attribute to ``file_name`` when saving.
        by default True
    **kwargs : Any
        Additional keyword arguments passes to the ``write_dataset`` implementation
        of the data io plugin. If you aren't sure about those use ``get_datasaver``
        to get the implementation with the proper help and autocomplete.
    """
    protect_from_overwrite(file_name, allow_overwrite=allow_overwrite)
    io = get_data_io(format_name or infer_file_format(file_name, needs_to_exist=False))
    if "loader" in dataset.attrs:
        del dataset.attrs["loader"]
    io.save_dataset(file_name=Path(file_name).as_posix(), dataset=dataset, **kwargs)
    dataset.attrs["loader"] = load_dataset
    if update_source_path is True or "source_path" not in dataset.attrs:
        dataset.attrs["source_path"] = Path(file_name).as_posix()


def get_dataloader(format_name: str) -> DataLoader:
    """Retrieve implementation of the ``read_dataset`` functionality for the format 'format_name'.

    This allows to get the proper help and autocomplete for the function,
    which is especially valuable if the function provides additional options.

    Parameters
    ----------
    format_name : str
        Format the dataloader should be able to read.

    Returns
    -------
    DataLoader
        Function to load data of format ``format_name`` as
        :xarraydoc:`Dataset` or :xarraydoc:`DataArray`.
    """
    io = get_data_io(format_name)
    return get_method_from_plugin(io, "load_dataset")


def get_datasaver(format_name: str) -> DataSaver:
    """Retrieve implementation of the ``save_dataset`` functionality for the format 'format_name'.

    This allows to get the proper help and autocomplete for the function,
    which is especially valuable if the function provides additional options.

    Parameters
    ----------
    format_name : str
        Format the datawriter should be able to write.

    Returns
    -------
    DataSaver
        Function to write :xarraydoc:`Dataset` to the format ``format_name`` .
    """
    io = get_data_io(format_name)
    return get_method_from_plugin(io, "save_dataset")


def show_data_io_method_help(
    format_name: str, method_name: Literal["load_dataset", "save_dataset"]
) -> None:
    """Show help for the implementation of data io plugin methods.

    Parameters
    ----------
    format_name : str
        Format the method should support.
    method_name : {'load_dataset', 'save_dataset'}
        Method name
    """
    io = get_data_io(format_name)
    show_method_help(io, method_name)


def data_io_plugin_table(*, plugin_names: bool = False, full_names: bool = False) -> MarkdownStr:
    """Return registered data io plugins and which functions they support as markdown table.

    This is especially useful when you work with new plugins.

    Parameters
    ----------
    plugin_names : bool
        Whether or not to add the names of the plugins to the table.
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    MarkdownStr
        Markdown table of data io plugins.
    """
    table_data = methods_differ_from_baseclass_table(
        DATA_IO_METHODS,
        known_data_formats(full_names=full_names),
        get_data_io,
        DataIoInterface,
        plugin_names=plugin_names,
    )
    header_values = ["Format name", *DATA_IO_METHODS]
    if plugin_names:
        header_values.append("Plugin name")
    headers = tuple(f"__{x}__" for x in header_values)
    return MarkdownStr(
        tabulate(
            bool_table_repr(table_data), tablefmt="github", headers=headers, stralign="center"
        )
    )


def supported_file_extensions_data_io(
    method_names: str | Sequence[str],
) -> Generator[str, None, None]:
    """Get data io formats that support all methods in ``method_names``.

    Parameters
    ----------
    method_names: str | Sequence[str]
        Names of Methods that need to support the file extension.

    Yields
    ------
    Generator[str, None, None]
        File extension supported by all methods in ``method_names``.

    See Also
    --------
    supported_file_extensions
    DATA_IO_METHODS
    """
    yield from supported_file_extensions(
        method_names,
        known_data_formats(),
        get_data_io,
        DataIoInterface,
    )



from __future__ import annotations

from dataclasses import replace
from functools import lru_cache
from pathlib import Path
from textwrap import dedent

import pytest
from pandas.testing import assert_frame_equal

from glotaran import __version__
from glotaran.io import load_result
from glotaran.io import save_dataset
from glotaran.io import save_result
from glotaran.optimization.optimize import optimize
from glotaran.project.result import Result
from glotaran.testing.simulated_data.sequential_spectral_decay import SCHEME
from glotaran.utils.io import chdir_context


@pytest.fixture
def dummy_result(tmp_path: Path):
    """Dummy result for testing."""

    @lru_cache
    def create_result():
        scheme = replace(SCHEME, maximum_number_function_evaluations=1)
        (tmp_path / "data").mkdir(parents=True, exist_ok=True)
        save_dataset(scheme.data["dataset_1"], tmp_path / "data/ds1.nc")
        return optimize(scheme, raise_exception=True)

    yield create_result()


@pytest.mark.parametrize("path_is_absolute", (True, False))
def test_save_result_yml(tmp_path: Path, dummy_result: Result, path_is_absolute: bool):
    """Check all files exist."""
    expected_result = dedent(
        f"""\
        number_of_function_evaluations: 1
        success: true
        termination_reason: The maximum number of function evaluations is exceeded.
        glotaran_version: {__version__}
        free_parameter_labels:
        - rates.species_1
        - rates.species_2
        - rates.species_3
        - irf.center
        - irf.width
        scheme: scheme.yml
        initial_parameters: initial_parameters.csv
        optimized_parameters: optimized_parameters.csv
        parameter_history: parameter_history.csv
        optimization_history: optimization_history.csv
        data:
          dataset_1: dataset_1.nc
        """
    )
    expected_scheme = dedent(
        """\
        model: model.yml
        parameters: initial_parameters.csv
        data:
          dataset_1: dataset_1.nc
        clp_link_tolerance: 0.0
        clp_link_method: nearest
        maximum_number_function_evaluations: 1
        add_svd: true
        ftol: 1e-08
        gtol: 1e-08
        xtol: 1e-08
        optimization_method: TrustRegionReflection
        result_path: null
        """
    )
    if path_is_absolute is True:
        result_dir = tmp_path / "testresult"
    else:
        result_dir = Path("testresult")

    assert (
        dummy_result.scheme.data["dataset_1"].source_path == (tmp_path / "data/ds1.nc").as_posix()
    )

    result_path = result_dir / "result.yml"
    with chdir_context("." if path_is_absolute is True else tmp_path):
        save_result(result_path=result_path, result=dummy_result)

        assert dummy_result.source_path == result_path.as_posix()

        assert (result_dir / "result.md").exists()
        assert (result_dir / "scheme.yml").exists()
        assert (result_dir / "scheme.yml").read_text() == expected_scheme
        assert result_path.exists()
        assert (result_dir / "initial_parameters.csv").exists()
        assert (result_dir / "optimized_parameters.csv").exists()
        assert (result_dir / "optimization_history.csv").exists()
        assert (result_dir / "dataset_1.nc").exists()
        # Original scheme object isn't changed
        assert (
            dummy_result.scheme.data["dataset_1"].source_path
            == (tmp_path / "data/ds1.nc").as_posix()
        )

        # We can't check equality due to numerical fluctuations
        got = result_path.read_text()
        print(got)
        assert expected_result in got


def test_save_result_yml_result_path_is_folder(tmp_path: Path, dummy_result: Result):
    """Save the result passing a folder instead of a file path"""
    result_folder = tmp_path / "testresult"
    save_result(result_path=result_folder, result=dummy_result)

    assert (result_folder / "result.yml").is_file()

    load_result(result_folder)


@pytest.mark.parametrize("path_is_absolute", (True, False))
def test_save_result_yml_roundtrip(tmp_path: Path, dummy_result: Result, path_is_absolute: bool):
    """Save and reloaded Result should be the same."""
    if path_is_absolute is True:
        result_dir = tmp_path / "testresult"
    else:
        result_dir = Path("testresult")
    result_path = result_dir / "result.yml"

    with chdir_context("." if path_is_absolute is True else tmp_path):
        save_result(result_path=result_path, result=dummy_result)
        result_round_tripped = load_result(result_path)

        assert dummy_result.source_path == result_path.as_posix()
        assert result_round_tripped.source_path == result_path.as_posix()

        assert_frame_equal(
            dummy_result.initial_parameters.to_dataframe(),
            result_round_tripped.initial_parameters.to_dataframe(),
        )
        assert_frame_equal(
            dummy_result.optimized_parameters.to_dataframe(),
            result_round_tripped.optimized_parameters.to_dataframe(),
        )
        assert_frame_equal(
            dummy_result.parameter_history.to_dataframe(),
            result_round_tripped.parameter_history.to_dataframe(),
        )
        assert_frame_equal(
            dummy_result.optimization_history.data, result_round_tripped.optimization_history.data
        )



"""Module for residual calculation with  the variable projection method."""
from __future__ import annotations

from typing import TYPE_CHECKING

from scipy.linalg import lapack

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


def residual_variable_projection(
    matrix: ArrayLike, data: ArrayLike
) -> tuple[ArrayLike, ArrayLike]:
    """Calculate conditionally linear parameters and residual with the variable projection method.

    Parameters
    ----------
    matrix : ArrayLike
        The model matrix.
    data : ArrayLike
        The data to analyze.

    Returns
    -------
    tuple[ArrayLike, ArrayLike]
        The clps and the residual.
    """
    # TODO: Reference Kaufman paper

    # Kaufman Q2 step 3
    qr, tau, _, _ = lapack.dgeqrf(matrix)

    # Kaufman Q2 step 4
    temp, _, _ = lapack.dormqr("L", "T", qr, tau, data, max(1, matrix.shape[1]), overwrite_c=0)

    clp, _ = lapack.dtrtrs(qr, temp)

    for i in range(matrix.shape[1]):
        temp[i] = 0

    # Kaufman Q2 step 5

    residual, _, _ = lapack.dormqr("L", "N", qr, tau, temp, max(1, matrix.shape[1]), overwrite_c=0)
    return clp[: matrix.shape[1]], residual



from __future__ import annotations

import importlib
from collections import namedtuple
from copy import deepcopy

import numpy as np

from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.io import prepare_time_trace_dataset
from glotaran.model import Model
from glotaran.optimization.optimize import optimize
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate

ParameterSpec = namedtuple("ParameterSpec", "base equal_area shapes")
NoiseSpec = namedtuple("NoiseSpec", "active seed std_dev")
SimulationSpec = namedtuple("SimulationSpec", "max_nfev noise")
DatasetSpec = namedtuple("DatasetSpec", "times wavelengths irf shapes")
IrfSpec = namedtuple("IrfSpec", "location width")
ShapeSpec = namedtuple("ShapeSpec", "amplitude location width")
ModelSpec = namedtuple("ModelSpec", "base shape spectral_megacomplex equ_area")
OptimizationSpec = namedtuple("OptimizationSpec", "nnls max_nfev")


SpectralDecayModel = Model.create_class_from_megacomplexes([DecayMegacomplex, SpectralMegacomplex])


def plot_overview(res, title=None):
    """very simple plot helper function derived from pyglotaran_extras"""
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(2, 2, figsize=(9, 8), constrained_layout=True)
    if title:
        fig.suptitle(title, fontsize=16)
    sas = res.species_associated_spectra
    traces = res.species_concentration
    if "spectral" in traces.coords:
        traces.sel(spectral=res.spectral[0], method="nearest").plot.line(x="time", ax=ax[0, 0])
    else:
        traces.plot.line(x="time", ax=ax[0, 0])
    sas.plot.line(x="spectral", ax=ax[0, 1])
    rLSV = res.residual_left_singular_vectors
    rLSV.isel(left_singular_value_index=range(min(2, len(rLSV)))).plot.line(x="time", ax=ax[1, 0])

    ax[1, 0].set_title("res. LSV")
    rRSV = res.residual_right_singular_vectors
    rRSV.isel(right_singular_value_index=range(min(2, len(rRSV)))).plot.line(
        x="spectral", ax=ax[1, 1]
    )

    ax[1, 1].set_title("res. RSV")
    plt.show(block=False)


def test_equal_area_penalties(debug=False):
    # %%

    optim_spec = OptimizationSpec(nnls=True, max_nfev=999)
    noise_spec = NoiseSpec(active=True, seed=1, std_dev=1e-8)

    wavelengths = np.arange(650, 670, 2)
    time_p1 = np.linspace(-1, 2, 50, endpoint=False)
    time_p2 = np.linspace(2, 10, 30, endpoint=False)
    time_p3 = np.geomspace(10, 50, num=20)
    times = np.concatenate([time_p1, time_p2, time_p3])

    irf_loc = float(times[20])
    irf_width = float((times[1] - times[0]) * 10)
    irf = IrfSpec(irf_loc, irf_width)

    amplitude = 1
    location1 = float(wavelengths[2])  # 2
    location2 = float(wavelengths[-3])  # -3
    width1 = float((wavelengths[1] - wavelengths[0]) * 5)
    width2 = float((wavelengths[1] - wavelengths[0]) * 3)
    shape1 = ShapeSpec(amplitude, location1, width1)
    shape2 = ShapeSpec(amplitude, location2, width2)
    dataset_spec = DatasetSpec(times, wavelengths, irf, [shape1, shape2])

    wavelengths = dataset_spec.wavelengths
    equ_interval = [(min(wavelengths), max(wavelengths))]
    weight = 0.01
    # %% The base model specification (mspec)
    base = {
        "initial_concentration": {
            "j1": {
                "compartments": ["s1", "s2"],
                "parameters": ["i.1", "i.2"],
            },
        },
        "megacomplex": {
            "mc1": {"type": "decay", "k_matrix": ["k1"]},
        },
        "k_matrix": {
            "k1": {
                "matrix": {
                    ("s1", "s1"): "kinetic.1",
                    ("s2", "s2"): "kinetic.2",
                }
            }
        },
        "irf": {
            "irf1": {"type": "gaussian", "center": "irf.center", "width": "irf.width"},
        },
        "dataset": {
            "dataset1": {
                "initial_concentration": "j1",
                "megacomplex": ["mc1"],
                "irf": "irf1",
            },
        },
    }

    shape = {
        "shape": {
            "sh1": {
                "type": "gaussian",
                "amplitude": "shapes.amps.1",
                "location": "shapes.locs.1",
                "width": "shapes.width.1",
            },
            "sh2": {
                "type": "gaussian",
                "amplitude": "shapes.amps.2",
                "location": "shapes.locs.2",
                "width": "shapes.width.2",
            },
        }
    }

    spectral_megacomplex = {
        "type": "spectral",
        "shape": {
            "s1": "sh1",
            "s2": "sh2",
        },
    }

    equ_area = {
        "clp_penalties": [
            {
                "type": "equal_area",
                "source": "s1",
                "target": "s2",
                "parameter": "rela.1",
                "source_intervals": equ_interval,
                "target_intervals": equ_interval,
                "weight": weight,
            },
        ],
    }
    mspec = ModelSpec(base, shape, spectral_megacomplex, equ_area)

    rela = 1.0  # relation between areas
    irf = dataset_spec.irf
    [sh1, sh2] = dataset_spec.shapes
    pspec_base = {
        "kinetic": [1e-1, 5e-3],
        "i": [0.5, 0.5, {"vary": False}],
        "irf": [["center", irf.location], ["width", irf.width]],
    }
    pspec_equa_area = {
        "rela": [rela, {"vary": False}],
    }
    pspec_shape = {
        "shapes": {
            "amps": [sh1.amplitude, sh2.amplitude],
            "locs": [sh1.location, sh2.location],
            "width": [sh1.width, sh2.width],
        },
    }
    pspec = ParameterSpec(pspec_base, pspec_equa_area, pspec_shape)

    # derivates:
    mspec_sim = dict(deepcopy(mspec.base), **mspec.shape)
    mspec_sim["megacomplex"]["mc2"] = mspec.spectral_megacomplex
    mspec_sim["dataset"]["dataset1"]["global_megacomplex"] = ["mc2"]

    mspec_fit_wp = dict(deepcopy(mspec.base), **mspec.equ_area)
    mspec_fit_np = dict(deepcopy(mspec.base))

    model_sim = SpectralDecayModel(**mspec_sim)
    model_wp = SpectralDecayModel(**mspec_fit_wp)
    model_np = SpectralDecayModel(**mspec_fit_np)
    print(model_np)

    # %% Parameter specification (pspec)

    pspec_sim = dict(deepcopy(pspec.base), **pspec.shapes)
    param_sim = Parameters.from_dict(pspec_sim)

    # For the wp model we create two version of the parameter specification
    # One has all inputs fixed, the other has all but the first free
    # for both we perturb kinetic parameters a bit to give the optimizer some work
    pspec_wp = dict(deepcopy(pspec.base), **pspec.equal_area)
    pspec_wp["kinetic"] = [v * 1.01 for v in pspec_wp["kinetic"]]
    pspec_wp["i"] = [[1, {"vary": False}], 1]
    pspec_np = dict(deepcopy(pspec.base))

    param_wp = Parameters.from_dict(pspec_wp)
    param_np = Parameters.from_dict(pspec_np)

    # %% Print models with parameters
    print(model_sim.markdown(param_sim))
    print(model_wp.markdown(param_wp))
    print(model_np.markdown(param_np))

    # %%
    simulated_data = simulate(
        model_sim,
        "dataset1",
        param_sim,
        coordinates={"time": times, "spectral": wavelengths},
        noise=noise_spec.active,
        noise_std_dev=noise_spec.std_dev,
        noise_seed=noise_spec.seed,
    )
    # %%
    simulated_data = prepare_time_trace_dataset(simulated_data)
    # make a copy to keep an intact reference
    data = deepcopy(simulated_data)

    # %% Optimizing model without penalty (np)

    model_np.dataset_groups["default"].method = (
        "non_negative_least_squares" if optim_spec.nnls else "variable_projection"
    )

    dataset = {"dataset1": data}
    scheme_np = Scheme(
        model=model_np,
        parameters=param_np,
        data=dataset,
        maximum_number_function_evaluations=optim_spec.max_nfev,
    )
    result_np = optimize(scheme_np, raise_exception=True)
    print(result_np)

    model_wp.dataset_groups["default"].method = (
        "non_negative_least_squares" if optim_spec.nnls else "variable_projection"
    )

    # %% Optimizing model with penalty fixed inputs (wp_ifix)
    scheme_wp = Scheme(
        model=model_wp,
        parameters=param_wp,
        data=dataset,
        maximum_number_function_evaluations=optim_spec.max_nfev,
    )
    result_wp = optimize(scheme_wp, raise_exception=True)
    print(result_wp)

    if debug:
        # %% Plot results
        plt_spec = importlib.util.find_spec("matplotlib")
        if plt_spec is not None:
            import matplotlib.pyplot as plt

            plot_overview(result_np.data["dataset1"], "no penalties")
            plot_overview(result_wp.data["dataset1"], "with penalties")
            plt.show()

    # %% Test calculation
    print(result_wp.data["dataset1"])
    area1_np = np.sum(result_np.data["dataset1"].species_associated_spectra.sel(species="s1"))
    area2_np = np.sum(result_np.data["dataset1"].species_associated_spectra.sel(species="s2"))
    print("area_np", area1_np, area2_np)
    assert not np.isclose(area1_np, area2_np)

    area1_wp = np.sum(result_wp.data["dataset1"].species_associated_spectra.sel(species="s1"))
    area2_wp = np.sum(result_wp.data["dataset1"].species_associated_spectra.sel(species="s2"))
    assert np.isclose(area1_wp, area2_wp)

    input_ratio = result_wp.optimized_parameters.get("i.1") / result_wp.optimized_parameters.get(
        "i.2"
    )
    print("input", input_ratio)
    assert np.isclose(input_ratio, 1.5038858115)


if __name__ == "__main__":
    test_equal_area_penalties(debug=True)
    test_equal_area_penalties(debug=False)



"""Project Io registration convenience functions.

Note
----
The [call-arg] type error would be raised since the base methods doesn't have a ``**kwargs``
argument, but we rather ignore this error here, than adding ``**kwargs`` to the base method
and causing an [override] type error in the plugins implementation.
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import TypeVar

from tabulate import tabulate

from glotaran.io.interface import SAVING_OPTIONS_DEFAULT
from glotaran.io.interface import ProjectIoInterface
from glotaran.io.interface import SavingOptions
from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.base_registry import add_instantiated_plugin_to_registry
from glotaran.plugin_system.base_registry import get_method_from_plugin
from glotaran.plugin_system.base_registry import get_plugin_from_registry
from glotaran.plugin_system.base_registry import is_registered_plugin
from glotaran.plugin_system.base_registry import methods_differ_from_baseclass_table
from glotaran.plugin_system.base_registry import registered_plugins
from glotaran.plugin_system.base_registry import set_plugin
from glotaran.plugin_system.base_registry import show_method_help
from glotaran.plugin_system.base_registry import supported_file_extensions
from glotaran.plugin_system.io_plugin_utils import bool_table_repr
from glotaran.plugin_system.io_plugin_utils import infer_file_format
from glotaran.plugin_system.io_plugin_utils import not_implemented_to_value_error
from glotaran.plugin_system.io_plugin_utils import protect_from_overwrite
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Generator
    from collections.abc import Sequence
    from typing import Any
    from typing import Literal

    from glotaran.model import Model
    from glotaran.parameter import Parameters
    from glotaran.project import Result
    from glotaran.project import Scheme
    from glotaran.typing import StrOrPath

    ProjectIoMethods = TypeVar(
        "ProjectIoMethods",
        Literal["load_model"],
        Literal["save_model"],
        Literal["load_parameters"],
        Literal["save_parameters"],
        Literal["load_scheme"],
        Literal["save_scheme"],
        Literal["load_result"],
        Literal["save_result"],
    )

PROJECT_IO_METHODS = (
    "load_model",
    "save_model",
    "load_parameters",
    "save_parameters",
    "load_scheme",
    "save_scheme",
    "load_result",
    "save_result",
)


def register_project_io(
    format_names: str | list[str],
) -> Callable[[type[ProjectIoInterface]], type[ProjectIoInterface]]:
    """Register project io plugins to one or more formats.

    Decorate a project io plugin class with ``@register_project_io(format_name|[*format_names])``
    to add it to the registry.

    Parameters
    ----------
    format_names : str | list[str]
        Name of the project io plugin under which it is registered.

    Returns
    -------
    Callable[[type[ProjectIoInterface]], type[ProjectIoInterface]]
        Inner decorator function.

    Examples
    --------
    >>> @register_project_io("my_format_1")
    ... class MyProjectIo1(ProjectIoInterface):
    ...     pass

    >>> @register_project_io(["my_format_1", "my_format_1_alias"])
    ... class MyProjectIo2(ProjectIoInterface):
    ...     pass
    """

    def wrapper(cls: type[ProjectIoInterface]) -> type[ProjectIoInterface]:
        add_instantiated_plugin_to_registry(
            plugin_register_keys=format_names,
            plugin_class=cls,
            plugin_registry=__PluginRegistry.project_io,
            plugin_set_func_name="set_project_plugin",
        )
        return cls

    return wrapper


def is_known_project_format(format_name: str) -> bool:
    """Check if a data format is in the project_io registry.

    Parameters
    ----------
    format_name : str
        Name of the project io plugin under which it is registered.

    Returns
    -------
    bool
        Whether or not the data format is a registered project io plugin.
    """
    return is_registered_plugin(
        plugin_register_key=format_name, plugin_registry=__PluginRegistry.project_io
    )


def known_project_formats(full_names: bool = False) -> list[str]:
    """Names of the registered project io plugins.

    Parameters
    ----------
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    list[str]
        List of registered project io plugins.
    """
    return registered_plugins(plugin_registry=__PluginRegistry.project_io, full_names=full_names)


def set_project_plugin(
    format_name: str,
    full_plugin_name: str,
) -> None:
    """Set the plugin used for a specific project format.

    This function is useful when you want to resolve conflicts of installed plugins
    or overwrite the plugin used for a specific format.

    Effected functions:

    - :func:`load_model`
    - :func:`save_model`
    - :func:`load_parameters`
    - :func:`save_parameters`
    - :func:`load_scheme`
    - :func:`save_scheme`
    - :func:`load_result`
    - :func:`save_result`

    Parameters
    ----------
    format_name : str
        Format name used to refer to the plugin when used for ``save`` and ``load`` functions.
    full_plugin_name : str
        Full name (import path) of the registered plugin.
    """
    set_plugin(
        plugin_register_key=format_name,
        full_plugin_name=full_plugin_name,
        plugin_registry=__PluginRegistry.project_io,
    )


def get_project_io(format_name: str) -> ProjectIoInterface:
    """Retrieve a data io plugin from the project_io registry.

    Parameters
    ----------
    format_name : str
        Name of the data io plugin under which it is registered.

    Returns
    -------
    ProjectIoInterface
        Project io plugin instance.
    """
    return get_plugin_from_registry(
        plugin_register_key=format_name,
        plugin_registry=__PluginRegistry.project_io,
        not_found_error_message=(
            f"Unknown Project Io format {format_name!r}. "
            f"Known formats are: {known_project_formats()}"
        ),
    )


@not_implemented_to_value_error
def load_model(file_name: StrOrPath, format_name: str | None = None, **kwargs: Any) -> Model:
    """Create a Model instance from the specs defined in a file.

    Parameters
    ----------
    file_name : StrOrPath
        File containing the model specs.
    format_name : str
        Format the file is in, if not provided it will be inferred from the file extension.
    **kwargs: Any
        Additional keyword arguments passes to the ``load_model`` implementation
        of the project io plugin.

    Returns
    -------
    Model
        Model instance created from the file.
    """
    io = get_project_io(format_name or infer_file_format(file_name))
    model = io.load_model(Path(file_name).as_posix(), **kwargs)
    model.source_path = Path(file_name).as_posix()
    return model


@not_implemented_to_value_error
def save_model(
    model: Model,
    file_name: StrOrPath,
    format_name: str | None = None,
    *,
    allow_overwrite: bool = False,
    update_source_path: bool = True,
    **kwargs: Any,
) -> None:
    """Save a :class:`Model` instance to a spec file.

    Parameters
    ----------
    model: Model
        :class:`Model` instance to save to specs file.
    file_name: StrOrPath
        File to write the model specs to.
    format_name: str | None
        Format the file should be in, if not provided it will be inferred from the file extension.
    allow_overwrite: bool
        Whether or not to allow overwriting existing files, by default False
    update_source_path: bool
        Whether or not to update the ``source_path`` attribute to ``file_name`` when saving.
        by default True
    **kwargs: Any
        Additional keyword arguments passes to the ``save_model`` implementation
        of the project io plugin.
    """
    protect_from_overwrite(file_name, allow_overwrite=allow_overwrite)
    io = get_project_io(format_name or infer_file_format(file_name, needs_to_exist=False))
    io.save_model(file_name=Path(file_name).as_posix(), model=model, **kwargs)
    if update_source_path is True:
        model.source_path = Path(file_name).as_posix()


@not_implemented_to_value_error
def load_parameters(file_name: StrOrPath, format_name: str | None = None, **kwargs) -> Parameters:
    """Create a :class:`Parameters` instance from the specs defined in a file.

    Parameters
    ----------
    file_name: StrOrPath
        File containing the parameter specs.
    format_name: str | None
        Format the file is in, if not provided it will be inferred from the file extension.
    **kwargs: Any
        Additional keyword arguments passes to the ``load_parameters`` implementation
        of the project io plugin.

    Returns
    -------
    Parameters
        :class:`Parameters` instance created from the file.

    .. # noqa: D414
    """
    io = get_project_io(format_name or infer_file_format(file_name))
    parameters = io.load_parameters(
        Path(file_name).as_posix(),
        **kwargs,
    )
    parameters.source_path = Path(file_name).as_posix()
    return parameters


@not_implemented_to_value_error
def save_parameters(
    parameters: Parameters,
    file_name: StrOrPath,
    format_name: str | None = None,
    *,
    allow_overwrite: bool = False,
    update_source_path: bool = True,
    **kwargs: Any,
) -> None:
    """Save a :class:`Parameters` instance to a spec file.

    Parameters
    ----------
    parameters: Parameters
        :class:`Parameters` instance to save to specs file.
    file_name: StrOrPath
        File to write the parameter specs to.
    format_name: str | None
        Format the file should be in, if not provided it will be inferred from the file extension.
    allow_overwrite: bool
        Whether or not to allow overwriting existing files, by default False
    update_source_path: bool
        Whether or not to update the ``source_path`` attribute to ``file_name`` when saving.
        by default True
    **kwargs: Any
        Additional keyword arguments passes to the ``save_parameters`` implementation
        of the project io plugin.
    """
    protect_from_overwrite(file_name, allow_overwrite=allow_overwrite)
    io = get_project_io(format_name or infer_file_format(file_name, needs_to_exist=False))
    io.save_parameters(file_name=Path(file_name).as_posix(), parameters=parameters, **kwargs)
    if update_source_path is True:
        parameters.source_path = Path(file_name).as_posix()


@not_implemented_to_value_error
def load_scheme(file_name: StrOrPath, format_name: str | None = None, **kwargs: Any) -> Scheme:
    """Create a :class:`Scheme` instance from the specs defined in a file.

    Parameters
    ----------
    file_name: StrOrPath
        File containing the parameter specs.
    format_name: str | None
        Format the file is in, if not provided it will be inferred from the file extension.
    **kwargs: Any
        Additional keyword arguments passes to the ``load_scheme`` implementation
        of the project io plugin.

    Returns
    -------
    Scheme
        :class:`Scheme` instance created from the file.
    """
    io = get_project_io(format_name or infer_file_format(file_name))

    scheme = io.load_scheme(Path(file_name).as_posix(), **kwargs)
    scheme.source_path = Path(file_name).as_posix()
    return scheme


@not_implemented_to_value_error
def save_scheme(
    scheme: Scheme,
    file_name: StrOrPath,
    format_name: str | None = None,
    *,
    allow_overwrite: bool = False,
    update_source_path: bool = True,
    **kwargs: Any,
) -> None:
    """Save a :class:`Scheme` instance to a spec file.

    Parameters
    ----------
    scheme: Scheme
        :class:`Scheme` instance to save to specs file.
    file_name: StrOrPath
        File to write the scheme specs to.
    format_name: str | None
        Format the file should be in, if not provided it will be inferred from the file extension.
    allow_overwrite: bool
        Whether or not to allow overwriting existing files, by default False
    update_source_path: bool
        Whether or not to update the ``source_path`` attribute to ``file_name`` when saving.
        by default True
    **kwargs: Any
        Additional keyword arguments passes to the ``save_scheme`` implementation
        of the project io plugin.
    """
    protect_from_overwrite(file_name, allow_overwrite=allow_overwrite)
    io = get_project_io(format_name or infer_file_format(file_name, needs_to_exist=False))
    io.save_scheme(file_name=Path(file_name).as_posix(), scheme=scheme, **kwargs)
    if update_source_path is True:
        scheme.source_path = Path(file_name).as_posix()


@not_implemented_to_value_error
def load_result(result_path: StrOrPath, format_name: str | None = None, **kwargs: Any) -> Result:
    """Create a :class:`Result` instance from the specs defined in a file.

    Parameters
    ----------
    result_path: StrOrPath
        Path containing the result data.
    format_name: str | None
        Format the result is in, if not provided and it is a file
        it will be inferred from the file extension.
    **kwargs: Any
        Additional keyword arguments passes to the ``load_result`` implementation
        of the project io plugin.

    Returns
    -------
    Result
        :class:`Result` instance created from the saved format.
    """
    io = get_project_io(format_name or infer_file_format(result_path, allow_folder=True))

    result = io.load_result(Path(result_path).as_posix(), **kwargs)
    result.source_path = Path(result_path).as_posix()
    return result


@not_implemented_to_value_error
def save_result(
    result: Result,
    result_path: StrOrPath,
    format_name: str | None = None,
    *,
    allow_overwrite: bool = False,
    update_source_path: bool = True,
    saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT,
    **kwargs: Any,
) -> list[str]:
    """Write a :class:`Result` instance to a spec file.

    Parameters
    ----------
    result: Result
        :class:`Result` instance to write.
    result_path: StrOrPath
        Path to write the result data to.
    format_name: str | None
        Format the result should be saved in, if not provided and it is a file
        it will be inferred from the file extension.
    allow_overwrite: bool
        Whether or not to allow overwriting existing files, by default False
    update_source_path: bool
        Whether or not to update the ``source_path`` attribute to ``result_path`` when saving.
        by default True
    saving_options: SavingOptions
        Options for the saved result.
    **kwargs: Any
        Additional keyword arguments passes to the ``save_result`` implementation
        of the project io plugin.

    Returns
    -------
    list[str] | None
        List of file paths which were saved.
    """
    protect_from_overwrite(result_path, allow_overwrite=allow_overwrite)
    io = get_project_io(
        format_name or infer_file_format(result_path, needs_to_exist=False, allow_folder=True)
    )
    paths = io.save_result(
        result_path=Path(result_path).as_posix(),
        result=result,
        saving_options=saving_options,
        **kwargs,
    )
    if update_source_path is True:
        result.source_path = Path(result_path).as_posix()
    return paths


def get_project_io_method(format_name: str, method_name: ProjectIoMethods) -> Callable[..., Any]:
    """Retrieve implementation of project io functionality for the format 'format_name'.

    This allows to get the proper help and autocomplete for the function,
    which is especially valuable if the function provides additional options.

    Parameters
    ----------
    format_name : str
        Format the dataloader should be able to read.
    method_name : {'load_model', 'write_model', 'load_parameters', 'write_parameters',\
    'load_scheme', 'write_scheme', 'load_result', 'write_result'}
        Method name, e.g. load_model.

    Returns
    -------
    Callable[..., Any]
        The function which is called in the background by the convenience functions.


    .. # noqa: DAR103 method_name
    """
    io = get_project_io(format_name)
    return get_method_from_plugin(io, method_name)


def show_project_io_method_help(format_name: str, method_name: ProjectIoMethods) -> None:
    """Show help for the implementation of project io plugin methods.

    Parameters
    ----------
    format_name : str
        Format the method should support.
    method_name : {'load_model', 'write_model', 'load_parameters', 'write_parameters',\
    'load_scheme', 'write_scheme', 'load_result', 'write_result'}
        Method name.


    .. # noqa: DAR103 method_name
    .. # noqa: DAR101
    """
    io = get_project_io(format_name)
    show_method_help(io, method_name)


def project_io_plugin_table(
    *, plugin_names: bool = False, full_names: bool = False
) -> MarkdownStr:
    """Return registered project io plugins and which functions they support as markdown table.

    This is especially useful when you work with new plugins.

    Parameters
    ----------
    plugin_names : bool
        Whether or not to add the names of the plugins to the table.
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    MarkdownStr
        Markdown table of project io plugins.
    """
    table_data = methods_differ_from_baseclass_table(
        PROJECT_IO_METHODS,
        known_project_formats(full_names=full_names),
        get_project_io,
        ProjectIoInterface,
        plugin_names=plugin_names,
    )
    header_values = ["Format name", *PROJECT_IO_METHODS]
    if plugin_names:
        header_values.append("Plugin name")
    headers = tuple(f"__{x}__" for x in header_values)
    return MarkdownStr(
        tabulate(
            bool_table_repr(table_data), tablefmt="github", headers=headers, stralign="center"
        )
    )


def supported_file_extensions_project_io(
    method_names: str | Sequence[str],
) -> Generator[str, None, None]:
    """Get project io formats that support all methods in ``method_names``.

    Parameters
    ----------
    method_names: str | Sequence[str]
        Names of Methods that need to support the file extension.

    Yields
    ------
    Generator[str, None, None]
        File extension supported by all methods in ``method_names``.

    See Also
    --------
    supported_file_extensions
    PROJECT_IO_METHODS
    """
    yield from supported_file_extensions(
        method_names,
        known_project_formats(),
        get_project_io,
        ProjectIoInterface,
    )



import numpy as np
import pytest
import xarray as xr

from glotaran.model import DatasetGroup
from glotaran.optimization.data_provider import DataProvider
from glotaran.optimization.data_provider import DataProviderLinked
from glotaran.optimization.test.models import SimpleTestModel
from glotaran.parameter import Parameters
from glotaran.project import Scheme


@pytest.fixture()
def dataset_one() -> xr.Dataset:
    global_axis = [1, 5, 6]
    model_axis = [5, 7, 9, 12]

    data = xr.DataArray(
        np.ones((4, 3)), coords=[("model", model_axis), ("global", global_axis)]
    ).to_dataset(name="data")

    data["weight"] = xr.ones_like(data.data) * 0.5
    return data


@pytest.fixture()
def dataset_two() -> xr.Dataset:
    global_axis = [0, 3, 7, 10]
    model_axis = [4, 11, 15]

    return xr.DataArray(
        np.ones((4, 3)) * 2, coords=[("global", global_axis), ("model", model_axis)]
    ).to_dataset(name="data")


@pytest.fixture()
def scheme(dataset_one: xr.Dataset, dataset_two: xr.Dataset) -> Scheme:
    model = SimpleTestModel(
        **{
            "megacomplex": {"m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                },
                "dataset2": {
                    "megacomplex": ["m1"],
                },
            },
        }
    )
    print(model.validate())
    assert model.valid()

    parameters = Parameters.from_list([])

    data = {"dataset1": dataset_one, "dataset2": dataset_two}
    return Scheme(model, parameters, data, clp_link_tolerance=1)


@pytest.fixture()
def dataset_group(scheme: Scheme) -> DatasetGroup:
    dataset_group = scheme.model.get_dataset_groups()["default"]
    dataset_group.set_parameters(scheme.parameters)
    return dataset_group


def test_data_provider(
    dataset_one: xr.Dataset, dataset_two: xr.Dataset, scheme: Scheme, dataset_group: DatasetGroup
):
    data_provider = DataProvider(scheme, dataset_group)

    print(dataset_one.data)
    print(data_provider.get_data("dataset1"))
    assert data_provider.get_model_dimension("dataset1") == "model"
    assert data_provider.get_global_dimension("dataset1") == "global"
    assert np.array_equal(
        dataset_one.data * dataset_one.weight, data_provider.get_data("dataset1")
    )
    assert np.array_equal(dataset_one.weight, data_provider.get_weight("dataset1"))
    assert np.array_equal(dataset_one.coords["model"], data_provider.get_model_axis("dataset1"))
    assert np.array_equal(dataset_one.coords["global"], data_provider.get_global_axis("dataset1"))

    assert data_provider.get_model_dimension("dataset2") == "model"
    assert data_provider.get_global_dimension("dataset2") == "global"
    assert np.array_equal(dataset_two.data.T, data_provider.get_data("dataset2"))
    assert data_provider.get_weight("dataset2") is None
    assert np.array_equal(dataset_two.coords["model"], data_provider.get_model_axis("dataset2"))
    assert np.array_equal(dataset_two.coords["global"], data_provider.get_global_axis("dataset2"))


def test_data_provider_linked(
    dataset_one: xr.Dataset, dataset_two: xr.Dataset, scheme: Scheme, dataset_group: DatasetGroup
):
    data_provider = DataProviderLinked(scheme, dataset_group)

    assert "dataset1" in data_provider.group_definitions
    assert data_provider.group_definitions["dataset1"] == ["dataset1"]
    assert "dataset2" in data_provider.group_definitions
    assert data_provider.group_definitions["dataset2"] == ["dataset2"]
    assert "dataset1dataset2" in data_provider.group_definitions
    assert data_provider.group_definitions["dataset1dataset2"] == ["dataset1", "dataset2"]

    #  global_axis1 = [1, 5, 6]
    #  global_axis2 = [0, 3, 7, 10]

    assert np.array_equal(data_provider.aligned_global_axis, [1, 3, 5, 6, 10])

    assert data_provider.get_aligned_group_label(0) == "dataset1dataset2"
    assert data_provider.get_aligned_group_label(1) == "dataset2"
    assert data_provider.get_aligned_group_label(2) == "dataset1"
    assert data_provider.get_aligned_group_label(3) == "dataset1dataset2"
    assert data_provider.get_aligned_group_label(4) == "dataset2"

    assert np.array_equal(data_provider.get_aligned_dataset_indices(0), [0, 0])
    assert np.array_equal(data_provider.get_aligned_dataset_indices(1), [1])
    assert np.array_equal(data_provider.get_aligned_dataset_indices(2), [1])
    assert np.array_equal(data_provider.get_aligned_dataset_indices(3), [2, 2])
    assert np.array_equal(data_provider.get_aligned_dataset_indices(4), [3])

    dataset1_size = dataset_one.coords["model"].size
    dataset2_size = dataset_two.coords["model"].size

    assert data_provider.get_aligned_data(0).size == dataset1_size + dataset2_size
    assert data_provider.get_aligned_data(1).size == dataset2_size
    assert data_provider.get_aligned_data(2).size == dataset1_size
    assert data_provider.get_aligned_data(3).size == dataset1_size + dataset2_size
    assert data_provider.get_aligned_data(4).size == dataset2_size

    assert (
        data_provider.get_aligned_weight(0).size  # type:ignore[union-attr]
        == dataset1_size + dataset2_size
    )
    assert data_provider.get_aligned_weight(1) is None
    assert data_provider.get_aligned_weight(2).size == dataset1_size  # type:ignore[union-attr]
    assert (
        data_provider.get_aligned_weight(3).size  # type:ignore[union-attr]
        == dataset1_size + dataset2_size
    )
    assert data_provider.get_aligned_weight(4) is None


@pytest.mark.parametrize("method", ["nearest", "backward", "forward"])
def test_data_provider_linking_methods(method: str, scheme: Scheme, dataset_group: DatasetGroup):
    scheme.clp_link_method = method  # type:ignore[assignment]
    data_provider = DataProviderLinked(scheme, dataset_group)

    #  global_axis1 = [1, 5, 6]
    #  global_axis2 = [0, 3, 7, 10]

    wanted_global_axis = [1, 3, 5, 6, 10]
    if method == "backward":
        wanted_global_axis = [0, 1, 3, 5, 6, 10]
    elif method == "forward":
        wanted_global_axis = [1, 3, 5, 6, 7, 10]
    assert np.array_equal(data_provider.aligned_global_axis, wanted_global_axis)



"""This module contains clp relation items."""
from __future__ import annotations

from glotaran.model.interval_item import IntervalItem
from glotaran.model.item import ParameterType
from glotaran.model.item import item


@item
class ClpRelation(IntervalItem):
    """Applies a relation between two clps.

    The relation is applied as :math:`target = parameter * source`.
    """

    source: str
    target: str
    parameter: ParameterType



"""This module contains the interval item."""
from __future__ import annotations

from glotaran.model.item import Item
from glotaran.model.item import item


@item
class IntervalItem(Item):
    """An item with an interval."""

    interval: tuple[float, float] | list[tuple[float, float]] | None = None

    def has_interval(self) -> bool:
        """Check if intervals are defined.

        Returns
        -------
        bool
        """
        return self.interval is not None

    def applies(self, index: float | None) -> bool:
        """Check if the index is in the intervals.

        Parameters
        ----------
        index : float
            The index.

        Returns
        -------
        bool

        """
        if self.interval is None or index is None:
            return True

        def applies(interval: tuple[float, float]):
            lower, upper = interval[0], interval[1]
            if lower > upper:
                lower, upper = upper, lower
            return lower <= index <= upper  # type:ignore[operator]

        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return any(applies(i) for i in self.interval)



import numpy as np

from glotaran.builtin.megacomplexes.clp_guide import ClpGuideMegacomplex
from glotaran.builtin.megacomplexes.decay import DecaySequentialMegacomplex
from glotaran.builtin.megacomplexes.decay.test.test_decay_megacomplex import create_gaussian_clp
from glotaran.model import Model
from glotaran.optimization.optimize import optimize
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation.simulation import simulate


def test_clp_guide():
    model = Model.create_class_from_megacomplexes(
        [DecaySequentialMegacomplex, ClpGuideMegacomplex]
    )(
        **{
            "dataset_groups": {"default": {"link_clp": True}},
            "megacomplex": {
                "mc1": {
                    "type": "decay-sequential",
                    "compartments": ["s1", "s2"],
                    "rates": ["1", "2"],
                },
                "mc2": {"type": "clp-guide", "dimension": "time", "target": "s1"},
            },
            "dataset": {
                "dataset1": {"megacomplex": ["mc1"]},
                "dataset2": {"megacomplex": ["mc2"]},
            },
        },
    )

    initial_parameters = Parameters.from_list(
        [101e-5, 501e-4, [1, {"vary": False, "non-negative": False}]]
    )
    wanted_parameters = Parameters.from_list(
        [101e-4, 501e-3, [1, {"vary": False, "non-negative": False}]]
    )

    time = np.arange(0, 50, 1.5)
    pixel = np.arange(600, 750, 5)
    axis = {"time": time, "pixel": pixel}

    clp = create_gaussian_clp(["s1", "s2"], [7, 30], [620, 720], [10, 50], pixel)

    dataset1 = simulate(model, "dataset1", wanted_parameters, axis, clp)
    dataset2 = clp.sel(clp_label=["s1"]).rename(clp_label="time")
    data = {"dataset1": dataset1, "dataset2": dataset2}

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme)
    print(result.optimized_parameters)
    for param in result.optimized_parameters.all():
        assert np.allclose(param.value, wanted_parameters.get(param.label).value, rtol=1e-1)



from glotaran.model.dataset_model import DatasetModel
from glotaran.model.item import ModelItem
from glotaran.model.item import ModelItemType
from glotaran.model.item import item
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import megacomplex
from glotaran.model.model import Model


@item
class MockItem(ModelItem):
    value: int


@item
class MockDatasetModel1(DatasetModel):
    test_dataset_prop: int


@megacomplex(dataset_model_type=MockDatasetModel1)
class MockMegacomplex1(Megacomplex):
    type: str = "mock-complex-1"
    test_item: ModelItemType[MockItem]


@item
class MockDatasetModel2(DatasetModel):
    test_dataset_str: str


@megacomplex(dataset_model_type=MockDatasetModel2)
class MockMegacomplex2(Megacomplex):
    type: str = "mock-complex-2"
    test_str: str = "foo"


def test_add_item_fields_to_model():
    mcls = Model.create_class_from_megacomplexes([MockMegacomplex1])
    m = mcls()
    print(m)
    assert isinstance(m.dataset, dict)
    assert m.dataset == {}
    assert isinstance(m.test_item, dict)
    assert m.test_item == {}

    m = mcls(
        dataset={"d1": {"megacomplex": ["m1"], "test_dataset_prop": 21}},
        megacomplex={"m1": {"type": "mock-complex-1", "test_item": "item1"}},
        test_item={"item1": {"value": 42}},
    )
    print(m)
    assert "m1" in m.megacomplex
    assert isinstance(m.megacomplex["m1"], MockMegacomplex1)
    assert m.megacomplex["m1"].test_item == "item1"

    assert "d1" in m.dataset
    assert m.dataset["d1"].label == "d1"
    assert m.dataset["d1"].megacomplex == ["m1"]
    assert m.dataset["d1"].test_dataset_prop == 21

    assert "item1" in m.test_item
    assert m.test_item["item1"].label == "item1"
    assert m.test_item["item1"].value == 42



"""Test deprecated imports from 'glotaran/__init__.py' """
from __future__ import annotations

from warnings import warn

import pytest

from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper


def dummy_warn(foo, bar=False):
    warn(GlotaranApiDeprecationWarning("foo"), stacklevel=2)
    if not isinstance(bar, bool):
        raise ValueError("not a bool")
    return foo, bar


def dummy_no_warn(foo, bar=False):
    return foo, bar


def test_deprecation_warning_on_call_test_helper():
    """Correct result passed on"""
    record, result = deprecation_warning_on_call_test_helper(
        dummy_warn, args=["foo"], kwargs={"bar": True}
    )
    assert len(record) == 1
    assert result == ("foo", True)


def test_deprecation_warning_on_call_test_helper_error_reraise():
    """Raise if raise_exception and args or kwargs"""

    with pytest.raises(ValueError, match="not a bool"):
        deprecation_warning_on_call_test_helper(
            dummy_warn, args=["foo"], kwargs={"bar": "baz"}, raise_exception=True
        )


@pytest.mark.xfail(strict=True, reason="Function did not warn.")
def test_deprecation_warning_on_call_test_helper_no_warn():
    """Fail no warning"""
    deprecation_warning_on_call_test_helper(dummy_no_warn, args=["foo"], kwargs={"bar": True})



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.spectral.shape import SpectralShape
from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import ModelItemType
from glotaran.model import item
from glotaran.model import megacomplex

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@item
class SpectralDatasetModel(DatasetModel):
    spectral_axis_inverted: bool = False
    spectral_axis_scale: float = 1


@megacomplex(dataset_model_type=SpectralDatasetModel)
class SpectralMegacomplex(Megacomplex):
    dimension: str = "spectral"
    type: str = "spectral"
    shape: dict[str, ModelItemType[SpectralShape]]

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        compartments = []
        for compartment in self.shape:
            if compartment in compartments:
                raise ModelError(f"More then one shape defined for compartment '{compartment}'")
            compartments.append(compartment)

        if dataset_model.spectral_axis_inverted:
            model_axis = dataset_model.spectral_axis_scale / model_axis
        elif dataset_model.spectral_axis_scale != 1:
            model_axis = model_axis * dataset_model.spectral_axis_scale

        dim1 = model_axis.size
        dim2 = len(self.shape)
        matrix = np.zeros((dim1, dim2))

        for i, shape in enumerate(self.shape.values()):
            matrix[:, i] += shape.calculate(model_axis)

        return compartments, matrix

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        species_dimension = "spectral_species" if as_global else "species"
        if species_dimension in dataset.coords:
            return

        species = []
        megacomplexes = (
            dataset_model.global_megacomplex if as_global else dataset_model.megacomplex
        )
        for m in megacomplexes:
            if isinstance(m, SpectralMegacomplex):
                species += [compartment for compartment in m.shape if compartment not in species]

        dataset.coords[species_dimension] = species
        matrix = dataset.global_matrix if as_global else dataset.matrix
        clp_dim = "global_clp_label" if as_global else "clp_label"
        dataset["species_spectra"] = (
            dataset.attrs["global_dimension"] if as_global else dataset.attrs["model_dimension"],
            species_dimension,
        ), matrix.sel({clp_dim: species}).values

        if not is_full_model:
            dataset["species_associated_concentrations"] = (
                (
                    dataset.attrs["global_dimension"],
                    species_dimension,
                ),
                dataset.clp.sel(clp_label=species).data,
            )



import os

DATA_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "data_files"))

LEGACY_FILES = {
    "flim_file": os.path.join(DATA_DIR, "FLIM.sdt.ascii"),
    "flim_traces": os.path.join(DATA_DIR, "FLIM_legacy_trace.ssv"),
    "flim_map": os.path.join(DATA_DIR, "FLIM_legacy_sum_map.ssv"),
}

TEMPORAL_DATA = {
    "sdt": os.path.join(DATA_DIR, "temporal.sdt"),
    "csv": os.path.join(DATA_DIR, "temporal.sdt.kin"),
}

FLIM_DATA = {
    "sdt": os.path.join(DATA_DIR, "FLIM.sdt"),
    "csv": os.path.join(DATA_DIR, "FLIM.sdt.ascii"),
}



from __future__ import annotations

import os.path
import re
from enum import Enum
from warnings import warn

import numpy as np
import pandas as pd
import xarray as xr

from glotaran.io import DataIoInterface
from glotaran.io import register_data_io
from glotaran.io.prepare_dataset import prepare_time_trace_dataset


class DataFileType(Enum):
    time_explicit = "Time explicit"
    wavelength_explicit = "Wavelength explicit"


class ExplicitFile:
    """
    Abstract class representing either a time- or wavelength-explicit file.
    """

    # TODO: implement time_intervals
    def __init__(self, filepath: str | None = None, dataset: xr.DataArray | None = None):
        self._file_data_format = None
        self._observations = []  # TODO: choose name: data_points, observations, data
        self._times = []
        self._spectral_indices = []
        self._label = ""
        self._comment = ""
        absfilepath = os.path.realpath(filepath)
        if dataset is not None:
            self._observations = np.array(dataset.values).T
            self._times = np.array(dataset.coords["time"])
            self._spectral_indices = np.array(dataset.coords["spectral"])
            self._file = filepath
        elif os.path.isfile(filepath):
            self._file = filepath
        elif os.path.isfile(absfilepath):
            self._file = absfilepath
        else:
            raise Exception(f"Path does not exist: {filepath}, {absfilepath}")

    def get_explicit_axis(self):
        raise NotImplementedError

    def set_explicit_axis(self, axis):
        raise NotImplementedError

    def get_secondary_axis(self):
        raise NotImplementedError

    def get_data_row(self, index):
        raise NotImplementedError

    def get_observations(self, index):
        raise NotImplementedError

    def get_format_name(self):
        raise NotImplementedError

    def write(
        self,
        overwrite=False,
        comment="",
        file_format=DataFileType.time_explicit,
        number_format="%.10e",
    ):
        # TODO: write a more elegant method

        if os.path.isfile(self._file) and not overwrite:
            raise FileExistsError(f"File already exist:\n{self._file}")
        comment = f"{self._comment} {comment}"

        comments = f"# Filename: {str(self._file)}\n{' '.join(comment.splitlines())}\n"

        if file_format == DataFileType.wavelength_explicit:
            wav = "\t".join(repr(num) for num in self._spectral_indices)
            header = (
                f"{comments}Wavelength explicit\nIntervalnr {len(self._spectral_indices)}\n{wav}"
            )
            raw_data = np.vstack((self._times.T, self._observations)).T
        elif file_format == DataFileType.time_explicit:
            tim = "\t".join(repr(num) for num in self._times)
            header = f"{comments}Time explicit\nIntervalnr {len(self._times)}\n{tim}"
            raw_data = np.vstack((self._spectral_indices.T, self._observations.T)).T
        else:
            raise NotImplementedError

        np.savetxt(
            self._file,
            raw_data,
            fmt=number_format,
            delimiter="\t",
            newline="\n",
            header=header,
            footer="",
            comments="",
        )

    def read(self, prepare: bool = True):
        if not os.path.isfile(self._file):
            raise FileNotFoundError("File does not exist.")
        with open(self._file) as f:
            f.readline()  # The first two lines are comments
            f.readline()
            # The third line defines the ExplicitFileFormat (Time or Wavelength explicit)
            self._file_data_format = get_data_file_format(f.readline())
            # The fourth line define the number of elements on the explicit axis, which
            # we can ignore because pandas is intelligent enough to read it
        # read the first line (explicit_axis) separately
        explicit_axis = pd.read_csv(
            self._file, skiprows=4, delimiter=r"\s+", header=None, nrows=1
        ).values
        explicit_axis = explicit_axis[0, :]  # reshape to (n,)
        # then the rest of the data:
        rest_of_data = pd.read_csv(self._file, skiprows=5, delimiter=r"\s+", header=None).values
        secondary_axis = rest_of_data[:, 0]
        observations = rest_of_data[:, 1:]
        if self._file_data_format == DataFileType.time_explicit:
            self._times = explicit_axis  # (501,)
            self._spectral_indices = secondary_axis  # (51,)
            self._observations = observations  # len(observation)=51 . (51, 501)
        elif self._file_data_format == DataFileType.wavelength_explicit:
            self._spectral_indices = explicit_axis
            self._times = secondary_axis
            self._observations = observations
        else:
            raise NotImplementedError()
        return self.dataset(prepare=prepare)

    def dataset(self, prepare: bool = True) -> xr.Dataset | xr.DataArray:
        data = self._observations
        if self._file_data_format == DataFileType.time_explicit:
            data = data.T
        dataset = xr.DataArray(
            data, coords=[("time", self._times), ("spectral", self._spectral_indices)]
        )
        if prepare:
            dataset = prepare_time_trace_dataset(dataset)
        return dataset


class WavelengthExplicitFile(ExplicitFile):
    """
    Represents a wavelength explicit file
    """

    def get_explicit_axis(self):
        return self._spectral_indices

    def get_secondary_axis(self):
        return self.observations()

    def get_data_row(self, index):
        return []

    def add_data_row(self, row):
        if self._timepoints is None:
            self._timepoints = []
        self._timepoints.append(float(row.pop(0)))

        if self._spectra is None:
            self._spectra = []
        self._spectra.append(float(row))

    def get_format_name(self):
        return DataFileType.wavelength_explicit

    def times(self):
        return self.get_secondary_axis()

    def wavelengths(self):
        return self.get_explicit_axis()


class TimeExplicitFile(ExplicitFile):
    """
    Represents a time explicit file
    """

    def get_explicit_axis(self):
        return self.observations()

    def set_explicit_axis(self, axies):
        self._timepoints = float(axies)

    def get_secondary_axis(self):
        return self.channel_labels

    def get_data_row(self, index):
        return self.get_channel(self.channel_labels()[index])

    def add_data_row(self, row):
        if self._spectral_indices is None:
            self._spectral_indices = []
        self._spectral_indices.append(row.pop(0))

        if self._spectra is None:
            self._spectra = []
        self._spectra.append(float(row))

    def get_format_name(self):
        return DataFileType.time_explicit


def get_interval_number(line):
    match = re.search(r"intervalnr\s(.*)", line.strip().lower())
    interval_number = match[1] if match else None
    if not interval_number:
        interval_number = re.search(r"\d+", line[::-1]).group()[::-1]
    try:
        interval_number = int(interval_number)
    except ValueError:
        warn(f"No interval number found in line:\n{line}")
        interval_number = None
    return interval_number


def get_data_file_format(line):
    data_file_format = None
    if re.search(r"time\s+explicit|time\t+explicit", line.strip().lower()):
        # print("Time explicit format") #TODO: verbosity / debug statement
        data_file_format = DataFileType.time_explicit
    elif re.search(r"wavelength\s+explicit|wavelength\t+explicit", line.strip().lower()):
        # print("Wavelength explicit format") #TODO: verbosity / debug statement
        data_file_format = DataFileType.wavelength_explicit
    else:
        raise NotImplementedError()
    return data_file_format


#  @file_reader(extension="ascii", name="Wavelength-/Time-Explicit ASCII")
@register_data_io("ascii")
class AsciiDataIo(DataIoInterface):
    def load_dataset(self, file_name: str, *, prepare: bool = True) -> xr.Dataset | xr.DataArray:
        """Reads an ascii file in wavelength- or time-explicit format.

        See [1]_ for documentation of this format.

        Parameters
        ----------
        fname : str
            Name of the ascii file.

        Returns
        -------
        dataset : xr.Dataset

        Notes
        -----
        .. [1] https://glotaran.github.io/legacy/file_formats
        """

        with open(file_name) as f:
            f.readline()  # Read first line with comments (and discard for now)
            f.readline()  # Read second line with comments (and discard for now)
            data_file_format = get_data_file_format(f.readline())

        data_file = (
            WavelengthExplicitFile(filepath=file_name)
            if data_file_format is DataFileType.wavelength_explicit
            else TimeExplicitFile(file_name)
        )

        return data_file.read(prepare=prepare)

    def save_dataset(
        self,
        dataset: xr.DataArray | xr.Dataset,
        file_name: str,
        *,
        comment: str = "",
        file_format: DataFileType = DataFileType.time_explicit,
        number_format: str = "%.10e",
    ):
        if isinstance(dataset, xr.Dataset) and "data" in dataset:
            dataset = dataset.data
            warn(
                UserWarning(
                    "Saving the 'data' attribute of 'dataset' as a fallback."
                    "Result saving for ascii format only supports xarray.DataArray format, "
                    "please pass a xarray.DataArray instead of a xarray.Dataset "
                    "(e.g. dataset.data)."
                ),
                stacklevel=4,
            )
        data_file = (
            TimeExplicitFile(filepath=file_name, dataset=dataset)
            if file_format is DataFileType.time_explicit
            else WavelengthExplicitFile(filepath=file_name, dataset=dataset)
        )
        data_file.write(
            overwrite=True, comment=comment, file_format=file_format, number_format=number_format
        )



"""Deprecation functions for the yaml parser."""
from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.deprecation import deprecate_dict_entry

if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from typing import Any


def model_spec_deprecations(spec: MutableMapping[Any, Any]) -> None:
    """Check deprecations in the model specification ``spec`` dict.

    Parameters
    ----------
    spec : MutableMapping[Any, Any]
        Model specification dictionary
    """
    load_model_stack_level = 7

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="clp_area_penalties",
        new_usage="clp_penalties",
        to_be_removed_in_version="0.8.0",
        swap_keys=("clp_area_penalties", "clp_penalties"),
        stacklevel=load_model_stack_level,
    )



"""The glotaran model package."""
from glotaran.model.clp_constraint import OnlyConstraint
from glotaran.model.clp_constraint import ZeroConstraint
from glotaran.model.clp_penalties import EqualAreaPenalty
from glotaran.model.clp_relation import ClpRelation
from glotaran.model.dataset_group import DatasetGroup
from glotaran.model.dataset_model import DatasetModel
from glotaran.model.dataset_model import get_dataset_model_model_dimension
from glotaran.model.item import ItemIssue
from glotaran.model.item import ModelItem
from glotaran.model.item import ModelItemType
from glotaran.model.item import ModelItemTyped
from glotaran.model.item import ParameterType
from glotaran.model.item import attribute
from glotaran.model.item import fill_item
from glotaran.model.item import item
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import megacomplex
from glotaran.model.model import Model
from glotaran.model.model import ModelError
from glotaran.model.weight import Weight



"""A simple parallel decay for testing purposes."""

from glotaran.io import load_model
from glotaran.project import Scheme
from glotaran.project.generators import generate_model_yml
from glotaran.simulation import simulate
from glotaran.testing.simulated_data.shared_decay import PARAMETERS
from glotaran.testing.simulated_data.shared_decay import SIMULATION_COORDINATES
from glotaran.testing.simulated_data.shared_decay import SIMULATION_PARAMETERS

SIMULATION_MODEL_YML = generate_model_yml(
    generator_name="spectral_decay_parallel",
    generator_arguments={"nr_compartments": 3, "irf": True},
)
SIMULATION_MODEL = load_model(SIMULATION_MODEL_YML, format_name="yml_str")

MODEL_YML = generate_model_yml(
    generator_name="decay_parallel",
    generator_arguments={"nr_compartments": 3, "irf": True},
)
MODEL = load_model(MODEL_YML, format_name="yml_str")

DATASET = simulate(
    SIMULATION_MODEL,
    "dataset_1",
    SIMULATION_PARAMETERS,
    SIMULATION_COORDINATES,
    noise=True,
    noise_std_dev=1e-2,
)

SCHEME = Scheme(model=MODEL, parameters=PARAMETERS, data={"dataset_1": DATASET})



from __future__ import annotations

import pickle
import re
from typing import Any

import numpy as np
import pytest

from glotaran.parameter import Parameter


@pytest.mark.parametrize(
    "key_name, value_1, value_2",
    (
        ("value", 1, 2),
        ("vary", True, False),
        ("minimum", -np.inf, -1),
        ("maximum", np.inf, 1),
        ("expression", None, "$a.1*10"),
        ("standard_error", np.nan, 1),
        ("non_negative", True, False),
    ),
)
def test_parameter__deep_equals(key_name: str, value_1: Any, value_2: Any):
    parameter_1 = Parameter(label="foo", **{key_name: value_1})
    parameter_2 = Parameter(label="foo", **{key_name: value_1})
    assert parameter_1._deep_equals(parameter_2)

    parameter_3 = Parameter(label="foo", **{key_name: value_1})
    parameter_4 = Parameter(label="foo", **{key_name: value_2})
    assert not parameter_3._deep_equals(parameter_4)


@pytest.mark.parametrize("label, expected", (("foo", "foo"), (0, "0"), (1, "1")))
def test_parameter_label_always_str_or_None(label: str | int, expected: str):
    """Parameter.label is always a string"""
    parameter = Parameter(label=label)
    assert parameter.label == expected


@pytest.mark.parametrize(
    "label",
    ("exp", np.nan),
)
def test_parameter_label_error_wrong_label_pattern(label: str | int | float):
    """Error if label can't be casted to a valid label str"""
    with pytest.raises(ValueError, match=f"'{label}' is not a valid parameter label."):
        Parameter(label=label)


@pytest.mark.parametrize(
    "parameter, expected_repr",
    (
        (
            Parameter(label="foo"),
            "Parameter(label='foo')",
        ),
        (
            Parameter(label="foo", expression="$foo.bar", value=1.0, vary=True),
            # vary gets set to False due to the usage of expression
            "Parameter(label='foo', value=1.0, expression='$foo.bar', vary=False)",
        ),
    ),
)
def test_parameter_repr(parameter: Parameter, expected_repr: str):
    """Repr creates code to recreate the object."""
    print(parameter.__repr__())
    assert parameter.__repr__() == expected_repr
    assert parameter._deep_equals(eval(expected_repr))


def test_parameter_from_list():
    params = [["5", 1], ["4", 2], ["3", 3]]

    parameters = [Parameter.from_list(v) for v in params]

    assert [p.label for p in parameters] == [v[0] for v in params]
    assert [p.value for p in parameters] == [v[1] for v in params]


def test_parameter_options():
    params = [
        ["5", 1, {"non-negative": False, "min": -1, "max": 1, "vary": False}],
        ["6", 4e2, {"non-negative": True, "min": -7e2, "max": 8e2, "vary": True}],
        ["7", 2e4],
    ]

    parameters = [Parameter.from_list(v) for v in params]

    assert parameters[0].value == 1.0
    assert not parameters[0].non_negative
    assert parameters[0].minimum == -1
    assert parameters[0].maximum == 1
    assert not parameters[0].vary

    assert parameters[1].value == 4e2
    assert parameters[1].non_negative
    assert parameters[1].minimum == -7e2
    assert parameters[1].maximum == 8e2
    assert parameters[1].vary

    assert parameters[2].value == 2e4
    assert not parameters[2].non_negative
    assert parameters[2].minimum == float("-inf")
    assert parameters[2].maximum == float("inf")
    assert parameters[2].vary


def test_parameter_value_not_numeric_error():
    """Error if value isn't numeric."""
    with pytest.raises(TypeError):
        Parameter(label="", value="foo")


def test_parameter_maximum_not_numeric_error():
    """Error if maximum isn't numeric."""
    with pytest.raises(TypeError):
        Parameter(label="", maximum="foo")


def test_parameter_minimum_not_numeric_error():
    """Error if minimum isn't numeric."""
    with pytest.raises(TypeError):
        Parameter(label="", minimum="foo")


def test_parameter_non_negative():
    notnonneg = Parameter(label="", value=1, non_negative=False)
    valuenotnoneg, _, _ = notnonneg.get_value_and_bounds_for_optimization()
    assert np.allclose(1, valuenotnoneg)
    notnonneg.set_value_from_optimization(valuenotnoneg)
    assert np.allclose(1, notnonneg.value)

    nonneg1 = Parameter(label="", value=1, non_negative=True)
    value1, _, _ = nonneg1.get_value_and_bounds_for_optimization()
    assert not np.allclose(1, value1)
    nonneg1.set_value_from_optimization(value1)
    assert np.allclose(1, nonneg1.value)

    nonneg2 = Parameter(label="", value=2, non_negative=True)
    value2, _, _ = nonneg2.get_value_and_bounds_for_optimization()
    assert not np.allclose(2, value2)
    nonneg2.set_value_from_optimization(value2)
    assert np.allclose(2, nonneg2.value)

    nonnegminmax = Parameter(label="", value=5, minimum=3, maximum=6, non_negative=True)
    value5, minimum, maximum = nonnegminmax.get_value_and_bounds_for_optimization()
    assert not np.allclose(5, value5)
    assert not np.allclose(3, minimum)
    assert not np.allclose(6, maximum)


@pytest.mark.parametrize(
    "case",
    [
        ("$1", "parameters.get('1').value"),
        (
            "1 - $kinetic.1 * exp($kinetic.2) + $kinetic.3",
            "1 - parameters.get('kinetic.1').value * exp(parameters.get('kinetic.2').value) "
            "+ parameters.get('kinetic.3').value",
        ),
        ("2", "2"),
        (
            "1 - sum([$kinetic.1, $kinetic.2])",
            "1 - sum([parameters.get('kinetic.1').value, parameters.get('kinetic.2').value])",
        ),
        ("exp($kinetic.4)", "exp(parameters.get('kinetic.4').value)"),
        ("$kinetic.5", "parameters.get('kinetic.5').value"),
        (
            "$parameters.parameters.param1 + $kinetic6",
            "parameters.get('parameters.parameters.param1').value "
            "+ parameters.get('kinetic6').value",
        ),
        (
            "$foo.7.bar + $kinetic6",
            "parameters.get('foo.7.bar').value " "+ parameters.get('kinetic6').value",
        ),
        ("$1", "parameters.get('1').value"),
        ("$1-$2", "parameters.get('1').value-parameters.get('2').value"),
        ("$1-$5", "parameters.get('1').value-parameters.get('5').value"),
        (
            "100 - $inputs1.s1 - $inputs1.s3 - $inputs1.s8 - $inputs1.s12",
            "100 - parameters.get('inputs1.s1').value - parameters.get('inputs1.s3').value "
            "- parameters.get('inputs1.s8').value - parameters.get('inputs1.s12').value",
        ),
    ],
)
def test_transform_expression(case):
    expression, wanted_parameters = case
    parameter = Parameter(label="", expression=expression)
    assert parameter.transformed_expression == wanted_parameters
    # just for the test coverage so the if condition is wrong
    assert parameter.transformed_expression == wanted_parameters


def test_label_validator():
    valid_names = [
        "1",
        "valid1",
        "_valid2",
        "extra_valid3",
    ]
    for label in valid_names:
        Parameter(label)

    invalid_names = [
        "testé",
        "kinetic::red",
        "kinetic_blue+kinetic_red",
        "makesthissense=trueandfalse",
        "what/about\\slashes",
        "$invalid",
        "round",
        "parameters",
    ]

    for label in invalid_names:
        print(label)
        with pytest.raises(
            ValueError, match=re.escape(f"'{label}' is not a valid parameter label.")
        ):
            Parameter(label=label)


def test_parameter_pickle(tmpdir):
    parameter = Parameter(
        label="testlabel",
        expression="testexpression",
        minimum=1,
        maximum=2,
        non_negative=True,
        value=42,
        vary=False,
    )

    with open(tmpdir.join("test_param_pickle"), "wb") as f:
        pickle.dump(parameter, f)
    with open(tmpdir.join("test_param_pickle"), "rb") as f:
        pickled_parameter = pickle.load(f)

    assert parameter == pickled_parameter


def test_parameter_numpy_operations():
    """Operators work like a float"""
    parm1 = Parameter(label="foo", value=1)
    parm1_neg = Parameter(label="foo", value=-1)
    parm2 = Parameter(label="foo", value=2)
    parm3 = Parameter(label="foo", value=3)
    parm3_5 = Parameter(label="foo", value=3.5)

    assert parm1 == 1
    assert parm1 != 2
    assert -parm1 == -1
    assert abs(parm1_neg) == 1
    assert int(parm1) == 1
    assert np.allclose(float(parm1), 1)
    assert np.allclose(parm1 + parm2, 3)
    assert np.allclose(parm3 - parm2, 1)
    assert np.allclose(parm3 * parm2, 6)
    assert np.allclose(parm3 / parm2, 1.5)
    assert parm3 // parm2 == 1
    assert np.trunc(parm3_5) == 3
    assert parm2 % 2 == 0
    assert 2 % parm2 == 0
    assert divmod(parm3, 2) == (1, 1)
    assert divmod(3, parm2) == (1, 1)
    assert np.allclose(parm3**parm2, 9)
    assert parm3 >= parm2
    assert parm3 > parm2
    assert parm1 <= parm2
    assert parm1 < parm2


def test_parameter_dict_roundtrip():
    param = Parameter(
        label="foo",
        expression="1",
        maximum=2,
        minimum=1,
        non_negative=True,
        value=42,
        vary=False,
    )

    param_dict = param.as_dict()
    print(param_dict)
    param_from_dict = Parameter(**param_dict)

    assert param.label == param_from_dict.label
    assert param.expression == param_from_dict.expression
    assert param.maximum == param_from_dict.maximum
    assert param.minimum == param_from_dict.minimum
    assert param.non_negative == param_from_dict.non_negative
    assert param.value == param_from_dict.value
    assert param.vary == param_from_dict.vary


def test_parameter_list_roundtrip():
    param = Parameter(
        label="foo",
        expression="1",
        maximum=2,
        minimum=1,
        non_negative=True,
        value=42,
        vary=False,
    )

    param_list = param.as_list()
    print(param_list)
    param_from_list = Parameter.from_list(param_list)

    assert param.label == param_from_list.label
    assert param.expression == param_from_list.expression
    assert param.maximum == param_from_list.maximum
    assert param.minimum == param_from_list.minimum
    assert param.non_negative == param_from_list.non_negative
    assert param.value == param_from_list.value
    assert param.vary == param_from_list.vary



"""This package contains irf items."""


import numpy as np

from glotaran.model import ModelError
from glotaran.model import ModelItemTyped
from glotaran.model import ParameterType
from glotaran.model import attribute
from glotaran.model import item


@item
class Irf(ModelItemTyped):
    """Represents an IRF."""


@item
class IrfMultiGaussian(Irf):
    """
    Represents a gaussian IRF.

    One width and one center is a single gauss.

    One center and multiple widths is a multiple gaussian.

    Multiple center and multiple widths is Double-, Triple- , etc. Gaussian.

    Parameters
    ----------

    label:
        label of the irf
    center:
        one or more center of the irf as parameter indices
    width:
        one or more widths of the gaussian as parameter index
    center_dispersion_coefficients:
        polynomial coefficients for the dispersion of the
        center as list of parameter indices. None for no dispersion.
    width_dispersion_coefficients:
        polynomial coefficients for the dispersion of the
        width as parameter indices. None for no dispersion.

    """

    type: str = "multi-gaussian"

    center: list[ParameterType]
    width: list[ParameterType]
    scale: list[ParameterType] | None = None
    shift: list[ParameterType] | None = None
    normalize: bool = True
    backsweep: bool = False
    backsweep_period: ParameterType | None = None

    def parameter(
        self, global_index: int, global_axis: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, float, bool, float]:
        """Returns the properties of the irf with shift applied."""

        centers = self.center if isinstance(self.center, list) else [self.center]
        centers = np.asarray([c.value for c in centers])

        widths = self.width if isinstance(self.width, list) else [self.width]
        widths = np.asarray([w.value for w in widths])

        len_centers = len(centers)
        len_widths = len(widths)
        if len_centers != len_widths:
            if min(len_centers, len_widths) != 1:
                raise ModelError(
                    f"len(centers) ({len_centers}) not equal "
                    f"len(widths) ({len_widths}) none of is 1."
                )
            if len_centers == 1:
                centers = np.asarray([centers[0] for _ in range(len_widths)])
            else:
                widths = np.asarray([widths[0] for _ in range(len_centers)])

        scales = self.scale if self.scale is not None else [1.0 for _ in centers]
        scales = scales if isinstance(scales, list) else [scales]
        scales = np.asarray(scales)

        shift = 0
        if self.shift is not None:
            if global_index >= len(self.shift):
                raise ModelError(
                    f"No shift parameter for index {global_index} "
                    f"({global_axis[global_index]}) in irf {self.label}"
                )
            shift = self.shift[global_index]

        backsweep = self.backsweep

        backsweep_period = self.backsweep_period.value if self.backsweep else 0

        return centers, widths, scales, shift, backsweep, backsweep_period

    def calculate(self, index: int, global_axis: np.ndarray, model_axis: np.ndarray) -> np.ndarray:
        centers, widths, scales, _, _, _ = self.parameter(index, global_axis)
        return sum(
            scale * np.exp(-1 * (model_axis - center) ** 2 / (2 * width**2))
            for center, width, scale in zip(centers, widths, scales)
        )

    def is_index_dependent(self):
        return self.shift is not None


@item
class IrfGaussian(IrfMultiGaussian):
    type: str = "gaussian"
    center: ParameterType
    width: ParameterType


@item
class IrfSpectralMultiGaussian(IrfMultiGaussian):
    """
    Represents a gaussian IRF.

    One width and one center is a single gauss.

    One center and multiple widths is a multiple gaussian.

    Multiple center and multiple widths is Double-, Triple- , etc. Gaussian.

    Parameters
    ----------

    label:
        label of the irf
    center:
        one or more center of the irf as parameter indices
    width:
        one or more widths of the gaussian as parameter index
    center_dispersion_coefficients:
        list of parameters with polynomial coefficients describing
        the dispersion of the irf center location. None for no dispersion.
    width_dispersion_coefficients:
        list of parameters with polynomial coefficients describing
        the dispersion of the width of the irf. None for no dispersion.

    """

    type: str = "spectral-multi-gaussian"
    dispersion_center: ParameterType
    center_dispersion_coefficients: list[ParameterType]
    width_dispersion_coefficients: list[ParameterType] = attribute(factory=list)
    model_dispersion_with_wavenumber: bool = False

    def parameter(self, global_index: int, global_axis: np.ndarray):
        """Returns the properties of the irf with shift and dispersion applied."""
        centers, widths, scale, shift, backsweep, backsweep_period = super().parameter(
            global_index, global_axis
        )

        index = global_axis[global_index] if global_index is not None else None

        if self.dispersion_center is not None:
            dist = (
                (1e3 / index - 1e3 / self.dispersion_center)
                if self.model_dispersion_with_wavenumber
                else (index - self.dispersion_center) / 100
            )

        if len(self.center_dispersion_coefficients) != 0:
            if self.dispersion_center is None:
                raise ModelError(f"No dispersion center defined for irf '{self.label}'")
            for i, disp in enumerate(self.center_dispersion_coefficients):
                centers += disp * np.power(dist, i + 1)

        if len(self.width_dispersion_coefficients) != 0:
            if self.dispersion_center is None:
                raise ModelError(f"No dispersion center defined for irf '{self.label}'")
            for i, disp in enumerate(self.width_dispersion_coefficients):
                widths = widths + disp * np.power(dist, i + 1)

        return centers, widths, scale, shift, backsweep, backsweep_period

    def calculate_dispersion(self, axis):
        dispersion = []
        for index, _ in enumerate(axis):
            center, _, _, _, _, _ = self.parameter(index, axis)
            dispersion.append(center)
        return np.asarray(dispersion).T

    def is_index_dependent(self):
        return super().is_index_dependent() or self.dispersion_center is not None


@item
class IrfSpectralGaussian(IrfSpectralMultiGaussian):
    type: str = "spectral-gaussian"
    center: ParameterType
    width: ParameterType



"""Module containing Excel like io support."""

from __future__ import annotations

import numpy as np
import pandas as pd

from glotaran.io import ProjectIoInterface
from glotaran.io import register_project_io
from glotaran.parameter import Parameters
from glotaran.parameter.parameter import OPTION_NAMES_DESERIALIZED
from glotaran.utils.io import safe_dataframe_fillna
from glotaran.utils.io import safe_dataframe_replace


@register_project_io(["xlsx", "ods"])
class ExcelProjectIo(ProjectIoInterface):
    """Plugin for Excel like data io."""

    def load_parameters(self, file_name: str) -> Parameters:
        """Load parameters from XLSX file.

        Parameters
        ----------
        file_name : str
            Name of file to be loaded.

        Returns
        -------
            :class:`Parameters`
        """
        df = pd.read_excel(file_name, na_values=["None", "none"])
        df.columns = [column.lower() for column in df.columns]
        df.rename(columns=OPTION_NAMES_DESERIALIZED, inplace=True)
        safe_dataframe_fillna(df, "minimum", -np.inf)
        safe_dataframe_fillna(df, "maximum", np.inf)
        return Parameters.from_dataframe(df, source=file_name)

    def save_parameters(self, parameters: Parameters, file_name: str):
        """Save a :class:`Parameters` to a Excel file.

        Parameters
        ----------
        parameters : Parameters
            Parameters to be saved to file.
        file_name : str
            File to write the parameters to.
        """
        df = parameters.to_dataframe()
        safe_dataframe_replace(df, "minimum", -np.inf, "")
        safe_dataframe_replace(df, "maximum", np.inf, "")
        df.to_excel(file_name, na_rep="None", index=False)



from attrs import fields

from glotaran.model.item import ModelItem
from glotaran.model.item import ModelItemType
from glotaran.model.item import ParameterType
from glotaran.model.item import fill_item
from glotaran.model.item import get_item_model_issues
from glotaran.model.item import get_item_parameter_issues
from glotaran.model.item import item
from glotaran.model.item import model_attributes
from glotaran.model.item import strip_type_and_structure_from_attribute
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import megacomplex
from glotaran.model.model import Model
from glotaran.parameter import Parameter
from glotaran.parameter import Parameters


@item
class MockModelItem(ModelItem):
    p_scalar: ParameterType
    p_list: list[ParameterType]
    p_dict: dict[str, ParameterType]


@megacomplex()
class MockMegacomplexItems(Megacomplex):
    type: str = "test_model_items_megacomplex"
    item1: ModelItemType[MockModelItem]
    item2: list[ModelItemType[MockModelItem]]
    item3: dict[str, ModelItemType[MockModelItem]]


def test_strip_type_and_structure_from_attribute():
    @item
    class MockItem:
        pscalar: int = None
        pscalar_option: int | None = None
        plist: list[int] = None
        plist_option: list[int] | None = None
        pdict: dict[str, int] = None
        pdict_option: dict[str, int] | None = None
        iscalar: ModelItemType[int] = None
        iscalar_option: ModelItemType[int] | None = None
        ilist: list[ModelItemType[int]] = None
        ilist_option: list[ModelItemType[int]] | None = None
        idict: dict[str, ModelItemType[int]] = None
        idict_option: dict[str, ModelItemType[int]] | None = None

    for attr in fields(MockItem):
        structure, type_ = strip_type_and_structure_from_attribute(attr)
        print(attr.name, attr.type, structure, type_)
        assert structure in (None, dict, list)
        assert type_ is int


def test_model_get_items():
    items = list(model_attributes(MockMegacomplexItems))

    assert len(items) == 3
    assert items[0].name == "item1"
    assert items[1].name == "item2"
    assert items[2].name == "item3"


def test_get_issues():
    mcls = Model.create_class_from_megacomplexes([MockMegacomplexItems])
    model = mcls(
        megacomplex={
            "m1": {
                "type": "test_model_items_megacomplex",
                "item1": "item1",
                "item2": ["item2"],
                "item3": {"foo": "item3"},
            }
        },
        item1={"test": {"p_scalar": "p1", "p_list": ["p2"], "p_dict": {"p": "p2"}}},
    )

    m = model.megacomplex["m1"]
    issues = get_item_model_issues(m, model)
    assert len(issues) == 3

    p = Parameters({})
    i = model.item1["test"]
    issues = get_item_parameter_issues(i, p)
    assert len(issues) == 3

    issues = model.get_issues(parameters=p)
    assert len(issues) == 6


def test_fill_item():
    mcls = Model.create_class_from_megacomplexes([MockMegacomplexItems])
    model = mcls(
        megacomplex={
            "m1": {
                "type": "test_model_items_megacomplex",
                "item1": "item",
                "item2": ["item"],
                "item3": {"foo": "item"},
            }
        },
        item1={"item": {"p_scalar": "1", "p_list": ["2"], "p_dict": {"p": "2"}}},
        item2={"item": {"p_scalar": "1", "p_list": ["2"], "p_dict": {"p": "2"}}},
        item3={"item": {"p_scalar": "1", "p_list": ["2"], "p_dict": {"p": "2"}}},
    )

    parameters = Parameters.from_list([2, 3, 4])
    assert model.valid(parameters)

    m = fill_item(model.megacomplex["m1"], model, parameters)
    assert isinstance(m.item1, MockModelItem)
    assert all(isinstance(v, MockModelItem) for v in m.item2)
    assert all(isinstance(v, MockModelItem) for v in m.item3.values())

    i = m.item1
    assert isinstance(i.p_scalar, Parameter)
    assert all(isinstance(v, Parameter) for v in i.p_list)
    assert all(isinstance(v, Parameter) for v in i.p_dict.values())
    assert i.p_scalar.value == 2
    assert i.p_list[0].value == 3
    assert i.p_dict["p"].value == 3



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import megacomplex

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@megacomplex(unique=True)
class BaselineMegacomplex(Megacomplex):
    type: str = "baseline"

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        clp_label = [f"{dataset_model.label}_baseline"]
        matrix = np.ones((model_axis.size, 1), dtype=np.float64)
        return clp_label, matrix

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        if not is_full_model:
            dataset["baseline"] = dataset.clp.sel(clp_label=f"{dataset_model.label}_baseline")



"""Module containing the ``OptimizationHistory`` class."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import pandas as pd

from glotaran.utils.regex import RegexPattern

if TYPE_CHECKING:
    from glotaran.typing import StrOrPath


class OptimizationHistory:
    """Wrapped DataFrame to hold information of the optimization and behaves like a ``DataFrame``.

    Ref.:
    https://stackoverflow.com/a/65375904/3990615
    """

    def __init__(self, data=None, source_path: StrOrPath | None = None) -> None:
        """Ensure DataFrame has the correct columns, is numeric and has iteration as index."""
        self._df = (
            pd.DataFrame(
                data,
                columns=["iteration", "nfev", "cost", "cost_reduction", "step_norm", "optimality"],
            )
            .apply(pd.to_numeric)
            .set_index("iteration")
        )
        if source_path is not None:
            self.source_path = Path(source_path).as_posix()
        else:
            self.source_path = "optimization_history.csv"

    def __getattr__(self, attr: str) -> Any:
        """Access class attribute and fallback to DataFrame attribute if not present.

        Parameters
        ----------
        attr: str
            Name of the attribute to access.

        Returns
        -------
        Any
            Attribute of ``OptimizationHistory`` or the DataFrame
        """
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self.data, attr)

    def __getitem__(self, column: str) -> pd.Series:
        """Access DataFrame instead of class items.

        Parameters
        ----------
        column: str
            Name of the column to access.

        Returns
        -------
        pd.Series
            Column of the DataFrame.
        """
        return self.data[column]

    @property
    def data(self) -> pd.DataFrame:
        """Underlying ``DataFrame`` which allows for autocomplete with static analyzers.

        Returns
        -------
        pd.DataFrame
            ``DataFrame`` containing ``OptimizationHistory`` data.
        """
        return self._df

    @classmethod
    def from_stdout_str(
        cls: type[OptimizationHistory], optimize_stdout: str
    ) -> OptimizationHistory:
        """Create ``OptimizationHistory`` instance from ``optimize_stdout``.

        Parameters
        ----------
        optimize_stdout: str
            SciPy optimization stdout string, read out via ``TeeContext.read()``.

        Returns
        -------
        OptimizationHistory
            ``OptimizationHistory`` instance created by parsing ``optimize_stdout``.
        """
        return cls(
            [m.groupdict() for m in RegexPattern.optimization_stdout.finditer(optimize_stdout)]
        )

    @classmethod
    def from_csv(cls: type[OptimizationHistory], path: StrOrPath) -> OptimizationHistory:
        """Read ``OptimizationHistory`` from file.

        Parameters
        ----------
        path : StrOrPath
            The path to the csv file.

        Returns
        -------
        OptimizationHistory
            ``OptimizationHistory`` read from file.
        """
        return cls(pd.read_csv(path), source_path=Path(path).as_posix())

    loader = from_csv

    def to_csv(self, path: StrOrPath, delimiter: str = ","):
        """Write a ``OptimizationHistory`` to a CSV file and set ``source_path``.

        Parameters
        ----------
        path : StrOrPath
            The path to the CSV file.
        delimiter : str
            The delimiter of the CSV file.
        """
        self.source_path = Path(path).as_posix()
        self.data.to_csv(path, sep=delimiter)



"""The glotaran registry module."""
from __future__ import annotations

from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from pathlib import Path
from typing import Any
from warnings import warn

from glotaran.utils.ipython import MarkdownStr


class AmbiguousNameWarning(UserWarning):
    """Warning thrown when an item with the same name already exists.

    This is the case if two files with the same name but different extensions exist next to
    each other.
    """

    def __init__(
        self,
        *,
        item_name: str,
        items: dict[str, Path],
        item_key: str,
        unique_item_key: str,
        project_root_path: Path,
    ):
        """Initialize ``AmbiguousNameWarning`` with a formatted message.

        Parameters
        ----------
        item_name: str
            Name of items in the registry (e.g. 'Parameters').
        items: dict[str, Path]
            Known items at this iteration point.
        item_key: str
            Key that would have been used if the file names weren't ambiguous.
        unique_item_key: str
            Unique key for the item with ambiguous file name.
        project_root_path: Path
            Root path of the project.
        """
        ambiguous_file_names = [
            value.relative_to(project_root_path).as_posix()
            for key, value in items.items()
            if key.startswith(item_key)
        ]
        super().__init__(
            f"The {item_name} name {item_key!r} is ambiguous since it could "
            f"refer to the following files: {ambiguous_file_names}\n"
            f"The file {items[unique_item_key].relative_to(project_root_path).as_posix()!r} "
            f"will be accessible by the name {unique_item_key!r}. \n"
            f"While {item_key!r} refers to the file "
            f"{items[item_key].relative_to(project_root_path).as_posix()!r}.\n"
            "Rename the files with unambiguous names to silence this warning."
        )


class ItemMapping(Mapping):
    """Container class for ``ProjectRegistry`` items.

    The main purpose of this class is to show a user friendly error when accessing none existing
    items.
    """

    def __init__(self, data: Mapping[str, Path], item_name: str) -> None:
        """Initialize class instance as wrapper around ``data``.

        Parameters
        ----------
        data: Mapping[str, Path]
            Underlying data that are used for mapping.
        item_name: str
            Name of items in the registry used to format warning and exception (e.g. 'Parameters').
        """
        self.data = dict(sorted(data.items()))
        self._item_name = item_name

    def __getitem__(self, key: str) -> Path:
        """Protocol method used when accessing an item."""
        if key in self.data:
            return self.data[key]
        raise ValueError(
            f"{self._item_name} '{key}' does not exist. "
            f"Known {self._item_name.rstrip('s')}s are: {list(self.data.keys())}"
        )

    def __iter__(self) -> Iterator[str]:
        """Protocol method used when iterating over an instance."""
        return iter(self.data)

    def __len__(self) -> int:
        """Protocol method used by ``len``."""
        return len(self.data)

    def __eq__(self, other: object) -> bool:
        """Protocol method used for equality checks."""
        if isinstance(other, ItemMapping):
            return self.data == other.data
        if isinstance(other, Mapping):  # sourcery skip: assign-if-exp
            return self.data == other
        return NotImplemented

    def __repr__(self) -> str:
        """Protocol method used to display instance."""
        return repr(self.data)


class ProjectRegistry:
    """A registry base class."""

    def __init__(
        self, directory: Path, file_suffix: str | Iterable[str], loader: Callable, item_name: str
    ):
        """Initialize a registry.

        Parameters
        ----------
        directory : Path
            The registry directory.
        file_suffix : str | Iterable[str]
            The suffixes of item files.
        loader : Callable
            A loader for the registry items.
        item_name: str
            Name of items in the registry used to format warning and exception (e.g. 'Parameters').
        """
        self._directory: Path = directory
        self._file_suffix: tuple[str, ...] = (
            (file_suffix,) if isinstance(file_suffix, str) else tuple(file_suffix)
        )
        self._loader: Callable = loader
        self._item_name = item_name

        self._create_directory_if_not_exist()

    @property
    def directory(self) -> Path:
        """Get the registry directory.

        Returns
        -------
        Path
            The registry directory.
        """
        return self._directory

    @property
    def empty(self) -> bool:
        """Whether the registry is empty.

        Returns
        -------
        bool
            Whether the registry is empty.
        """
        return len(self.items) == 0

    @property
    def items(self) -> ItemMapping:
        """Get the items of the registry.

        Returns
        -------
        ItemMapping
            The items of the registry.
        """
        items = {}
        for path in sorted(self._directory.rglob("*")):
            if self.is_item(path) is True:
                rel_parent_path = path.parent.relative_to(self._directory)
                item_key = (rel_parent_path / path.stem).as_posix()
                if item_key not in items:
                    items[item_key] = path
                else:
                    unique_item_key = (rel_parent_path / path.name).as_posix()
                    items[unique_item_key] = path
                    warn(
                        AmbiguousNameWarning(
                            item_name=self._item_name,
                            items=items,
                            item_key=item_key,
                            unique_item_key=unique_item_key,
                            project_root_path=self._directory.parent,
                        ),
                        stacklevel=3,
                    )
        return ItemMapping(items, self._item_name)

    def is_item(self, path: Path) -> bool:
        """Check if the path contains an registry item.

        Parameters
        ----------
        path : Path
            The path to check.

        Returns
        -------
        bool :
            Whether the path contains an item.
        """
        return path.suffix in self._file_suffix

    def load_item(self, name: str) -> Any:
        """Load an registry item by it's name.

        Parameters
        ----------
        name : str
            The item name.

        Returns
        -------
        Any
            The loaded item.

        Raises
        ------
        ValueError
            Raise if the item does not exist.


        .. # noqa: DAR402
        """
        return self._loader(self.items[name])

    def markdown(self, join_indentation: int = 0) -> MarkdownStr:
        """Format the registry items as a markdown text.

        Parameters
        ----------
        join_indentation: int
            Number of whitespaces to indent when joining the parts.
            This is intended to be used with dedent when used in an indented f-string.
            Defaults to 0.

        Returns
        -------
        MarkdownStr : str
            The markdown string.
        """
        if self.empty:
            return MarkdownStr("_None_")

        join_str = " " * join_indentation
        md = join_str.join(f"* {name}\n" for name in self.items)
        return MarkdownStr(md)

    def _create_directory_if_not_exist(self):
        """Create the registry directory if it does not exist."""
        self._directory.mkdir(parents=True, exist_ok=True)



import pytest
import xarray as xr

from glotaran.io.preprocessor import PreProcessingPipeline


def test_correct_baseline_value():
    pl = PreProcessingPipeline().correct_baseline_value(1)
    data = xr.DataArray([[1]])
    result = pl.apply(data)
    assert result == data - 1


@pytest.mark.parametrize("indexer", (slice(0, 2), [0, 1]))
def test_correct_baseline_average(indexer: slice | list[int]):
    pl = PreProcessingPipeline().correct_baseline_average(select={"dim_0": 0, "dim_1": indexer})
    data = xr.DataArray([[1.1, 0.9]])
    result = pl.apply(data)
    assert (result == data - 1).all()


def test_correct_baseline_average_exclude():
    pl = PreProcessingPipeline().correct_baseline_average(
        select={"dim_0": 0}, exclude={"dim_1": 1}
    )
    data = xr.DataArray([[1.1, 0.9]])
    result = pl.apply(data)
    print(result)
    assert (result == data - 1.1).all()


def test_to_from_dict():
    pl = (
        PreProcessingPipeline()
        .correct_baseline_value(1)
        .correct_baseline_average({"dim_1": slice(0, 2)})
    )
    pl_dict = pl.dict()
    assert pl_dict == {
        "actions": [
            {"action": "baseline-value", "value": 1.0},
            {"action": "baseline-average", "select": {"dim_1": slice(0, 2)}, "exclude": None},
        ]
    }
    assert PreProcessingPipeline.parse_obj(pl_dict) == pl



"""Megacomplex registration convenience functions."""
from __future__ import annotations

from typing import TYPE_CHECKING

from tabulate import tabulate

from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.base_registry import add_plugin_to_registry
from glotaran.plugin_system.base_registry import full_plugin_name
from glotaran.plugin_system.base_registry import get_plugin_from_registry
from glotaran.plugin_system.base_registry import is_registered_plugin
from glotaran.plugin_system.base_registry import registered_plugins
from glotaran.plugin_system.base_registry import set_plugin
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from glotaran.model import Megacomplex


def register_megacomplex(megacomplex_type: str, megacomplex: type[Megacomplex]) -> None:
    """Add a megacomplex to the megacomplex registry.

    Parameters
    ----------
    megacomplex_type : str
        Name of the megacomplex under which it is registered.
    megacomplex : type[Megacomplex]
        megacomplex class to be registered.
    """
    add_plugin_to_registry(
        plugin_register_key=megacomplex_type,
        plugin=megacomplex,
        plugin_registry=__PluginRegistry.megacomplex,
        plugin_set_func_name="set_megacomplex_plugin",
    )


def is_known_megacomplex(megacomplex_type: str) -> bool:
    """Check if a megacomplex is in the megacomplex registry.

    Parameters
    ----------
    megacomplex_type : str
        Name of the megacomplex under which it is registered.

    Returns
    -------
    bool
        Whether or not the megacomplex is registered.
    """
    return is_registered_plugin(
        plugin_register_key=megacomplex_type, plugin_registry=__PluginRegistry.megacomplex
    )


def get_megacomplex(megacomplex_type: str) -> type[Megacomplex]:
    """Retrieve a megacomplex from the megacomplex registry.

    Parameters
    ----------
    megacomplex_type : str
        Name of the megacomplex under which it is registered.

    Returns
    -------
    type[Megacomplex]
        Megacomplex class
    """
    return get_plugin_from_registry(
        plugin_register_key=megacomplex_type,
        plugin_registry=__PluginRegistry.megacomplex,
        not_found_error_message=(
            f"Unknown megacomplex type {megacomplex_type!r}. "
            f"Known megacomplex types are: {known_megacomplex_names(full_names=True)}"
        ),
    )


def known_megacomplex_names(full_names: bool = False) -> list[str]:
    """Names of the registered megacomplexs.

    Parameters
    ----------
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    list[str]
        List of registered megacomplexs.
    """
    return registered_plugins(__PluginRegistry.megacomplex, full_names=full_names)


def set_megacomplex_plugin(megacomplex_name: str, full_plugin_name: str) -> None:
    """Set the plugin used for a specific megacomplex name.

    This function is useful when you want to resolve conflicts of installed plugins
    or overwrite the plugin used for a specific megacomplex name.

    Effected functions:

    - :func:`optimize`

    Parameters
    ----------
    megacomplex_name : str
        Name of the megacomplex to use the plugin for.
    full_plugin_name : str
        Full name (import path) of the registered plugin.
    """
    set_plugin(
        plugin_register_key=megacomplex_name,
        full_plugin_name=full_plugin_name,
        plugin_registry=__PluginRegistry.megacomplex,
        plugin_register_key_name="megacomplex_name",
    )


def megacomplex_plugin_table(
    *, plugin_names: bool = False, full_names: bool = False
) -> MarkdownStr:
    """Return registered megacomplex plugins as markdown table.

    This is especially useful when you work with new plugins.

    Parameters
    ----------
    plugin_names : bool
        Whether or not to add the names of the plugins to the table.
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    MarkdownStr
        Markdown table of megacomplexnames.
    """
    table_data: list[list[str]] = []
    megacomplex_names = known_megacomplex_names(full_names=full_names)
    header_values = ["Megacomplex name"]
    if plugin_names:
        header_values.append("Plugin name")
        table_data.extend(
            [
                f"`{megacomplex_name}`",
                f"`{full_plugin_name(get_megacomplex(megacomplex_name))}`",
            ]
            for megacomplex_name in megacomplex_names
        )

    else:
        table_data = [[f"`{megacomplex_name}`"] for megacomplex_name in megacomplex_names]
    headers = tuple(f"__{x}__" for x in header_values)
    return MarkdownStr(tabulate(table_data, tablefmt="github", headers=headers, stralign="center"))



from typing import Any

import numpy as np
import pytest
import xarray as xr

from glotaran.builtin.megacomplexes.coherent_artifact import CoherentArtifactMegacomplex
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.model import Model
from glotaran.model import fill_item
from glotaran.optimization.matrix_provider import MatrixProvider
from glotaran.optimization.optimize import optimize
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate


@pytest.mark.parametrize(
    "spectral_dependence",
    ("none", "dispersed", "shifted"),
)
def test_coherent_artifact(spectral_dependence: str):
    model_dict: dict[str, dict[str, Any]] = {
        "initial_concentration": {
            "j1": {"compartments": ["s1"], "parameters": ["irf_center"]},
        },
        "megacomplex": {
            "mc1": {"type": "decay", "k_matrix": ["k1"]},
            "mc2": {"type": "coherent-artifact", "order": 3},
        },
        "k_matrix": {
            "k1": {
                "matrix": {
                    ("s1", "s1"): "rate",
                }
            }
        },
        "irf": {
            "irf1": {
                "type": "multi-gaussian",
                "center": ["irf_center"],
                "width": ["irf_width"],
            },
        },
        "dataset": {
            "dataset1": {
                "initial_concentration": "j1",
                "megacomplex": ["mc1", "mc2"],
                "irf": "irf1",
            },
        },
    }

    parameter_list = [
        ["rate", 101e-4],
        ["irf_center", 10, {"vary": False, "non-negative": False}],
        ["irf_width", 20, {"vary": False, "non-negative": False}],
    ]

    irf_spec = model_dict["irf"]["irf1"]

    if spectral_dependence == "dispersed":
        irf_spec["type"] = "spectral-multi-gaussian"
        irf_spec["dispersion_center"] = "irf_dispc"
        irf_spec["center_dispersion_coefficients"] = ["irf_disp1", "irf_disp2"]

        parameter_list += [
            ["irf_dispc", 300, {"vary": False, "non-negative": False}],
            ["irf_disp1", 0.01, {"vary": False, "non-negative": False}],
            ["irf_disp2", 0.001, {"vary": False, "non-negative": False}],
        ]
    elif spectral_dependence == "shifted":
        irf_spec["shift"] = ["irf_shift1", "irf_shift2", "irf_shift3"]
        parameter_list += [
            ["irf_shift1", -2],
            ["irf_shift2", 0],
            ["irf_shift3", 2],
        ]

    model = Model.create_class_from_megacomplexes([DecayMegacomplex, CoherentArtifactMegacomplex])(
        **model_dict
    )

    parameters = Parameters.from_list(parameter_list)  # type: ignore[arg-type]

    time = np.arange(0, 50, 1.5)
    spectral = np.asarray([200, 300, 400])

    dataset_model = fill_item(model.dataset["dataset1"], model, parameters)
    matrix = MatrixProvider.calculate_dataset_matrix(dataset_model, spectral, time)
    compartments = matrix.clp_labels

    print(compartments)
    assert len(compartments) == 4
    for i in range(1, 4):
        assert compartments[i] == f"coherent_artifact_{i}"

    if spectral_dependence == "none":
        assert matrix.matrix.shape == (time.size, 4)
    else:
        assert matrix.matrix.shape == (spectral.size, time.size, 4)

    clp = xr.DataArray(
        np.ones((3, 4)),
        coords=[
            ("spectral", spectral),
            (
                "clp_label",
                [
                    "s1",
                    "coherent_artifact_1",
                    "coherent_artifact_2",
                    "coherent_artifact_3",
                ],
            ),
        ],
    )
    axis = {"time": time, "spectral": clp.spectral.data}
    data = simulate(model, "dataset1", parameters, axis, clp)

    dataset = {"dataset1": data}
    scheme = Scheme(
        model=model, parameters=parameters, data=dataset, maximum_number_function_evaluations=20
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for param in result.optimized_parameters.all():
        assert np.allclose(param.value, parameters.get(param.label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(data.time, resultdata.time)
    assert np.array_equal(data.spectral, resultdata.spectral)
    assert data.data.shape == resultdata.data.shape
    assert data.data.shape == resultdata.fitted_data.shape
    assert np.allclose(data.data, resultdata.fitted_data)

    assert "coherent_artifact_response" in resultdata
    if spectral_dependence == "none":
        assert resultdata["coherent_artifact_response"].shape == (time.size, 3)
    else:
        assert resultdata["coherent_artifact_response"].shape == (spectral.size, time.size, 3)

    assert "coherent_artifact_associated_spectra" in resultdata
    assert resultdata["coherent_artifact_associated_spectra"].shape == (3, 3)



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from glotaran.io import load_parameters
from glotaran.io import save_parameters
from glotaran.parameter import Parameters

PANDAS_TEST_DATA = Path(__file__).parent / "data"
PATH_XLSX = PANDAS_TEST_DATA / "reference_parameters.xlsx"
PATH_ODS = PANDAS_TEST_DATA / "reference_parameters.ods"
PATH_CSV = PANDAS_TEST_DATA / "reference_parameters.csv"
PATH_TSV = PANDAS_TEST_DATA / "reference_parameters.tsv"
PATH_XLSX_ALT = PANDAS_TEST_DATA / "reference_parameters_alternative_notation.xlsx"
PATH_CSV_ALT = PANDAS_TEST_DATA / "reference_parameters_alternative_notation.csv"
PATH_CSV_SUBSET = PANDAS_TEST_DATA / "reference_parameters_subset.csv"


@pytest.fixture(scope="module")
def yaml_reference() -> Parameters:
    """Fixture for yaml reference data."""
    return load_parameters(PANDAS_TEST_DATA / "reference_parameters.yaml")


@pytest.fixture(scope="module")
def yaml_reference_subset() -> Parameters:
    """Fixture for yaml subset reference data."""
    return load_parameters(PANDAS_TEST_DATA / "reference_parameters_subset.yaml")


@pytest.mark.parametrize("reference_path", (PATH_XLSX, PATH_ODS, PATH_CSV, PATH_TSV))
def test_references(yaml_reference: Parameters, reference_path: Path):
    """References are the same"""
    result = load_parameters(reference_path)
    assert result == yaml_reference


def test_alternative_notations(yaml_reference: Parameters):
    """Reading parameter file with alternate syntax works and is case insensitive."""
    assert load_parameters(PATH_CSV_ALT) == yaml_reference
    assert load_parameters(PATH_XLSX_ALT) == yaml_reference


def test_csv_subset_notations(yaml_reference_subset: Parameters):
    """Reading parameter file with a susbset of possible attributes works."""
    assert load_parameters(PATH_CSV_SUBSET) == yaml_reference_subset


@pytest.mark.parametrize(
    "format_name,reference_path",
    (("xlsx", PATH_XLSX), ("ods", PATH_ODS), ("csv", PATH_CSV), ("tsv", PATH_TSV)),
)
def test_roundtrips(
    yaml_reference: Parameters, tmp_path: Path, format_name: str, reference_path: Path
):
    """Roundtrip via save and load have the same data."""
    format_reference = load_parameters(reference_path)
    parameter_path = tmp_path / f"test_parameters.{format_name}"
    save_parameters(file_name=parameter_path, format_name=format_name, parameters=yaml_reference)
    parameters_roundtrip = load_parameters(parameter_path)

    assert parameters_roundtrip == yaml_reference
    assert parameters_roundtrip == format_reference

    if format_name in {"csv", "tsv"}:
        assert parameter_path.read_text() == reference_path.read_text()

        first_data_line = parameter_path.read_text().splitlines()[1]
        sep = "," if format_name == "csv" else "\t"

        assert f"{sep}-inf" not in first_data_line
        assert f"{sep}inf" not in first_data_line
    else:
        assert_frame_equal(
            pd.read_excel(parameter_path, na_values=["None", "none"]),
            pd.read_excel(reference_path, na_values=["None", "none"]),
        )


@pytest.mark.parametrize("format_name,sep", (("csv", ","), ("tsv", "\t")))
def test_replace_infinfinity(
    yaml_reference: Parameters, tmp_path: Path, format_name: str, sep: str
):
    parameter_path = tmp_path / f"test_parameters.{format_name}"
    save_parameters(
        file_name=parameter_path,
        format_name=format_name,
        parameters=yaml_reference,
        replace_infinfinity=False,
    )
    df = pd.read_csv(parameter_path, sep=sep)
    df = df[df["label"] != "verbose_list.no_defaults"]
    assert all(df["maximum"] == np.inf)
    assert all(df["minimum"] == -np.inf)

    first_data_line = parameter_path.read_text().splitlines()[1]
    assert f"{sep}-inf" in first_data_line
    assert f"{sep}inf" in first_data_line

    assert load_parameters(parameter_path) == yaml_reference
